/*-------------------------------------------------------------------------
--
--  access_ip_db.c - functions to extract database entries from the IP
--  address database.
--
--  notes:
--
===========================================================================
--                           UPDATE HISTORY                              --
--                                                                       --
-- VER     Date         Explanation of Changes              Author       --
---------------------------------------------------------------------------
--  0   |  7/30/92 | Initial version                 | Steve Miller - USC   
--      |          |                                 |
-------------------------------------------------------------------------*/
#include <stdio.h>
#include <malloc.h>
#include <sys/types.h>
#include "checker.h"
 
 
u_long ip_hash();
extern FILE *mfp;
extern FILE *ofp;
extern IP_DATA *ip_tbl[IP_HASH_TABLE_SIZE];
extern u_long current_time;
extern u_long total_ip_records;
extern u_long total_er_records;
extern u_long ck_debug_flag;
extern u_long error_threshold;

 
 
/*------------------------------------------------------------------------
--
--  get_ip_record() - A routine to find a record in the ip address
--  database.
--
--  parameters:
--      ipaddr - IP address that serves as key into the database.
--
--  return value:
--      NULL if a matching record is not found,
--      pointer to a matching record if one is found.
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
IP_DATA *get_ip_record( ipaddr )
u_long ipaddr;
    {
    u_long hcode;
    IP_DATA *ipptr;
 
    /* Create hash of IP address given as database key. */
 
    hcode = ip_hash( ipaddr );
 
    /* Loop through the data records at this location in the table,
       if a record with a matching key is found then return a pointer
       to it. */
 
    for( ipptr=ip_tbl[hcode]; ipptr != NULL; ipptr=ipptr->flink )
         if( ipptr->remote_ip_addr == ipaddr ) return( ipptr );       
 
    /* If we did not find a matching key then return NULL */
 
    return( NULL );
    }

 
 
/*------------------------------------------------------------------------
--
--  create_ip_record() - A routine to create a record in the ip address
--  database if one doesn't exist.
--
--  parameters:
--      ipaddr - IP address that serves as key into the database.
--
--  return value:
--      pointer to the newly created record, or an already existing
--          record if one is found, or NULL if there is no more
--          dynamic memory available for the calloc() call.
--
--  notes:
--      This routine assumes a matching record does not already
--      exist (ie this call must be protected by an earlier
--      get_ip_record() call.
--
------------------------------------------------------------------------*/
IP_DATA *create_ip_record( ipaddr )
u_long ipaddr;
    {
    IP_DATA *tptr;
    u_long hcode;
 
    /* Create hash of IP address given as database key. */
 
    hcode = ip_hash( ipaddr );
 
    /* Create an empty record, if there is an error then just
       return the NULL pointer and let the caller deal with it. */
 
    tptr = (IP_DATA *)calloc( 1, sizeof(IP_DATA) );
    if( tptr == NULL ) return( tptr );
    if(mfp) fprintf( mfp, "allocated %X to %X for IP_DATA\n",
        tptr, tptr+sizeof(IP_DATA)-1 );
 
    /* Attach the existing contents of the hash table location to
       the new record's forward link. */
 
    tptr->flink = ip_tbl[hcode];
    /* Place this record's address in the hash table location. */
 
    ip_tbl[hcode] = tptr;
 
    /* Move the input IP address into the record's key loaction. */
 
    tptr->remote_ip_addr = ipaddr;
    total_ip_records++;
    return( tptr );
    }
 
 
 
/*------------------------------------------------------------------------
--
--  update_ip_record() - A routine to extract data from a query message
--  and place the data into the IP address database.
--
--  parameters:
--      rptr -  a pointer to the IP_DATA structure to be updated
--      qp   -  a pointer to the query message packet used as input
--
--  return value:
--      none.
--
--  notes:
--      none.
--
------------------------------------------------------------------------*/
void update_ip_record( rptr, type )
IP_DATA *rptr;
u_long type;
    {
    int i;
 
    /* If the incoming message is a query then update the state, and
       set up the time fields. */
 
    if( type != RESPONSE ) {
        rptr->previous_state = rptr->state;
        rptr->state = WAITING_FOR_RESPONSE;
        rptr->q_previous_time = rptr->q_latest_time;
        rptr->q_latest_time = current_time;
        if( rptr->q_first_time == 0 )
            rptr->q_first_time = current_time;
 
        /* handle the leaky bucket algorithm */
 
        if( rptr->q_lbucket_count == MAX_IPQ_STACK ) i = 1;
        else i = 0;
 
        for( ; i<rptr->q_lbucket_count; i++ )
            if( current_time-IPQ_LBKT_WINDOW < rptr->q_stack[i] ) break;
 
        if( i > 0 ) { memcpy( rptr->q_stack,
            &(rptr->q_stack[i]), sizeof(rptr->q_stack) - i*sizeof(u_long) );
            rptr->q_lbucket_count -= i; }
 
        rptr->q_stack[rptr->q_lbucket_count] = current_time;
        rptr->q_lbucket_count++;
 
        }
 
    /* If the incoming message is a response then just change the
       state and zero out the leaky bucket stack. */
 
    else {
        rptr->previous_state = rptr->state;
        rptr->state = RESPONSE_RECEIVED;
        rptr->q_lbucket_count = 0;
        memset( rptr->q_stack, 0, sizeof(rptr->q_stack) );
        }
    }

 
 
/*------------------------------------------------------------------------
--
--  delete_ip_record() - A routine to remove an ip database record after
--  all of the query/response records referencing it have been deleted.
--
--  parameters:
--      rptr -  a pointer to the IP_DATA structure to be updated
--
--  return value:
--      none.
--
--  notes:
--      none.
--
------------------------------------------------------------------------*/
void delete_ip_record( rptr )
IP_DATA *rptr;
    {
    IP_DATA **tptr;
    u_long hcode;
 
    /* Make sure we dump the error record before we get rid of the
       record. */
 
    dump_err_record( ofp, rptr, error_threshold );
 
    /* Get the hash code for the address being worked. */
 
    hcode = ip_hash(rptr->remote_ip_addr);
 
    /* Loop down the queue that starts at the hash table looking
       for a record that points to the record to be deleted. */
 
    for( tptr = &ip_tbl[hcode]; *tptr != NULL; tptr = &(*tptr)->flink )
        if( *tptr == rptr ) break;
    if( *tptr != NULL ) *tptr = rptr->flink;
 
    /* Check to see if an error record had been allocated.  If so then
       free it up before doing the main ip address record. */
 
    if( rptr->err_data != NULL ) {
        free( rptr->err_data );
        if(mfp) fprintf( mfp, "freed ERR_DATA at %X\n", rptr->err_data );
        total_er_records--; }
 
    /* Check to see if a saved error record is allocated.  If so then
       free it up before doing the main ip address record. */
 
    if( rptr->pec_save != NULL ) {
        free( rptr->pec_save );
        if(mfp) fprintf( mfp, "freed ERR_DATA at %X\n", rptr->pec_save );
        total_er_records--; }
 
    /* Check to see if the record had ever had its address resolved into
       a name, and free the space if necessary. */
 
    if( rptr->resolved_name ) {
        free( rptr->resolved_name );
        if(mfp) fprintf( mfp,"freed name string at %X\n",rptr->resolved_name);
        }
 
    /* Free up the space the record was occupying. */
 
    free( rptr );
    if(mfp) fprintf( mfp, "freed IP_DATA at %X\n", rptr );
    total_ip_records--;
    }

 
 
/*------------------------------------------------------------------------
--
--  ip_hash - a simple little routine to create a hash code from an
--  IP internet address.
--
--  parameters:
--      ipaddr - the input address value.
--
--  return value:
--      The hash code.
--
--  notes:
--      none.
--
------------------------------------------------------------------------*/
u_long ip_hash( ipaddr )
u_long ipaddr;
    {
 
    /* Create the hash code by just mod'ing with the size of the
       hash code table. */
 
    return( ipaddr % IP_HASH_TABLE_SIZE );
    }
