/*-------------------------------------------------------------------------
--
--  access_qr_db.c - functions to extract database entries from the
--  query/response database.
--
--  notes:
--
===========================================================================
--                           UPDATE HISTORY                              --
--                                                                       --
-- VER     Date         Explanation of Changes              Author       --
---------------------------------------------------------------------------
--  0   |  7/30/92 | Initial version                 | Steve Miller - USC
--  1   |  9/20/92 | Added IPD record handling       | Steve Miller - USC  
--      |          |                                 |
-------------------------------------------------------------------------*/
#include <stdio.h>
#include <malloc.h>
#include <sys/types.h>
#include <arpa/nameser.h>
#include "checker.h"
 
extern FILE *mfp;
extern QR_DATA  timeq[TIME_QUEUE_SLOT_COUNT];
extern QR_HTABLE *qr_tbl[QR_HASH_TABLE_SIZE];
extern u_long current_time;
extern u_long total_qr_records;
extern u_long total_ht_records;
extern u_long total_dm_records;
extern u_long total_qn_chars;
extern u_long total_queries;
extern u_long total_responses;
extern u_long ck_debug_flag;
 
extern int countvl;
 
void delete_ip_record();
u_long what_class();
u_long qr_hash();
int  find_ourname();
IPD_DATA *find_ipd_record();
IPD_DATA *create_ipd_record();

 
 
/*------------------------------------------------------------------------
--
--  get_qr_record() - A routine to find a record in the query/response
--  database.
--
--  parameters:
--      qr_name - Query name that serves as part of key into the database.
--      qr_type - Query type that serves as other part of key
--
--  return value:
--      NULL if a matching record is not found,
--      pointer to a matching record if one is found.
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
QR_DATA *get_qr_record( qr_name, qr_type, ipptr )
char *qr_name;
u_long qr_type;
IP_DATA *ipptr;
    {
    int i;
    QR_HTABLE *hp;
    QR_DATA *qp;
    u_long hcode;

    /* If the query name is not already lower case then change it. */

    for( i=0; i<strlen(qr_name); i++ ) qr_name[i] = tolower(qr_name[i]);
 
    /* Hash the query name and type, and look to see if can find a
       table entry with matching keys. */
 
    hcode = qr_hash( qr_name, qr_type );
    for( hp=qr_tbl[hcode]; hp != NULL; hp=hp->hclink ) {
        if( strcmp( hp->q_name, qr_name ) == 0 &&
            hp->q_type == qr_type ) break;
        }
 
    /* Return NULL if there was no matching hash table entry. */
 
    if( hp == NULL ) return( NULL );
 
    /* Otherwise look down the list of QR_DATA entries to see if any
       of them have the same IP address as the record we are looking
       for. */
 
    for( qp=hp->qrlink; qp != NULL; qp=qp->qrk.flink )
        if( qp->qrk.remote_ip_addr == ipptr->remote_ip_addr ) break;
 
    /* Return whatever we found to the caller. */
 
    return( qp );
    }

 
 
/*------------------------------------------------------------------------
--
--  create_qr_record() - A routine to create a record in the
--  query/response database.
--
--  parameters:
--      qr_name - Query name that serves as part of key into the database
--      qr_type - Query type that serves as other part of key
--      ipptr   - Pointer to the IP addr DB record for this query's addr
--
--  return value:
--      pointer to the newly created record
--
--  notes:
--      none.
--
------------------------------------------------------------------------*/
QR_DATA *create_qr_record( qr_name, qr_type, ipptr )
char *qr_name;
u_long qr_type;
IP_DATA *ipptr;
    {
    int i;
    static int slot=0;
    QR_HTABLE *hp;
    QR_DATA   *qrpt;
    QR_HTABLE **thp;
    IPD_DATA *ipdp;
    u_int hcode;
    char temp[200];

    /* If the query name is not already lower case then change it. */

    for( i=0; i<strlen(qr_name); i++ ) qr_name[i] = tolower(qr_name[i]);
 
    /* Allocate the space for the hash table entry and the query
       record, if there is not enough space return a NULL pointer to
       the caller as a signal that we had memory problems. */
 
    qrpt = (QR_DATA *)calloc( 1, sizeof(QR_DATA) );
    if( qrpt == NULL ) return( NULL );
    if(mfp) fprintf( mfp, "allocated %X to %X for QR_DATA\n",
       qrpt, qrpt+sizeof(QR_DATA)-1 );
 
    /* Hash the query name and type, and look to see if we need
       to create a new hash table entry. */
 
    hcode = qr_hash( qr_name, qr_type );
    for( hp=qr_tbl[hcode]; hp != NULL; hp=hp->hclink )
        if( strcmp( hp->q_name, qr_name ) == 0 &&
            hp->q_type == qr_type ) break;
 
    /* If we could not find a hash table entry that matches our keys
       then create one, place it at the front of the list, and fill
       out the key fields in the table entry structure. */
 
    if( hp == NULL ) {
 
        /* Allocate a new hash table entry, return a null pointer if
           we have problems getting the memory. */
 
        hp = (QR_HTABLE *)calloc( 1, sizeof(QR_HTABLE) );
        if( hp == NULL ) {
            free( qrpt );
            return( NULL ); }
        if(mfp) fprintf( mfp, "allocated %X to %X for QR_HTABLE (hcode=%d)\n",
           hp, hp+sizeof(QR_HTABLE)-1, hcode );
 
        /* Allocate space for the query name, but if there is not enough
           dynamic memory left to do that then clean up the things we
           have already allocated and return a null pointer. */
 
        hp->q_name = (char *)calloc( 1, strlen(qr_name)+1 );
        if( hp->q_name == NULL ) {
            free( qrpt );
            free( hp );
            return( NULL ); }
        if(mfp) fprintf( mfp, "allocated %X to %X for qr string\n",
           hp->q_name, hp->q_name+strlen(qr_name) );
 
        /* Knit together the records and buffers that we just allocated. */
 
        hp->hclink = qr_tbl[hcode];
        qr_tbl[hcode] = hp;
        strcpy( hp->q_name, qr_name );
        hp->q_type = qr_type;
        hp->qrlink = qrpt;
        total_qn_chars += strlen(qr_name) + 1;
        total_ht_records++;
        }
 
    /* If we did find an existing hash table entry, then insert this
       new record at the front of the list. */
 
    else {
        qrpt->qrk.flink = hp->qrlink;
        hp->qrlink = qrpt;
        }
 
    /* Insert the new query record into the current deletion queue
       slot and cycle the slot number. */
 
    qrpt->qrk.tq_flink = timeq[slot].qrk.tq_flink;
    timeq[slot].qrk.tq_flink = qrpt;
    qrpt->qrk.tq_blink = &timeq[slot];
    slot = (slot + 1) % TIME_QUEUE_SLOT_COUNT;
    total_qr_records++;
 
    /* Fill in the query record fields */
 
    qrpt->qrk.iprec = ipptr;
    qrpt->qrk.remote_ip_addr = ipptr->remote_ip_addr;
    qrpt->qrk.last_access_time = current_time;
    qrpt->qrk.blink = hp;
    qrpt->qrq.q_first_time = current_time;
    ipptr->reference_count++;
 
    /* Look for an IPD record, create one if it doesn't exist, and
       then update its reference count. */
 
    ipdp = find_ipd_record( qrpt->qrk.remote_ip_addr, hp );
    if( ipdp == NULL )
        ipdp = create_ipd_record( qrpt->qrk.remote_ip_addr, hp );
    if( ipdp != NULL ) {
        ipdp->reference_count++;
        qrpt->qrk.ipd_link = ipdp; }
 
    /* Return the new query record's pointer to the caller. */
 
    return( qrpt ); 
    }
 
 
 
/*------------------------------------------------------------------------
--
--  update_qr_record() - A routine to extract data from a query message
--  and place the data into the query/response database.
--
--  parameters:
--      qr -  a pointer to the QR_DATA structure to be updated
--      qp -  a pointer to the query message packet used as input
--
--  return value:
--      none.
--
--  notes:
--      none.
--
------------------------------------------------------------------------*/
void update_qr_record( qr, qp, sp )
QR_DATA *qr;
QUERY_PACKET *qp;
SOURCE_INFO *sp;
    {
    int i;
    IPD_DATA *ipdp;
 
    /* If this is a query then fill out the query fields in the
       QR_DATA structure. */
 
    if( qp->type == QUERY_MESSAGE ) {
 
        qr->qrk.last_access_time = current_time;
        qr->qrs.previous_state = qr->qrs.state;
        qr->qrs.state = WAITING_FOR_RESPONSE;
        qr->qrq.q_count++;
        qr->qrq.q_previous_id = qr->qrq.q_id;
        qr->qrq.q_previous_time = qr->qrq.q_latest_time;
        qr->qrq.q_latest_time = current_time;
        qr->qrq.q_id = qp->id;
        qr->qrq.q_port = sp->portnum;
        qr->qrq.q_protocol = sp->protocol;
        total_queries++;
        }
 
 
    /* If this is a response to a query then fill out the response
       related fields of the QR_DATA structure. */
 
    else if( qp->type == RESPONSE_MESSAGE ) {
 
        qr->qrs.previous_state = qr->qrs.state;
        qr->qrs.state = RESPONSE_RECEIVED;
        qr->qrr.r_previous_time = qr->qrr.r_latest_time;
        qr->qrr.r_latest_time = current_time;
        qr->qrr.r_count++;
        total_responses++;
 
        /* Move in the response parameters from the message. */
 
        qr->qrr.r_aa = qp->aa;
        qr->qrr.r_ttl = qp->ttl;
        qr->qrr.r_rcode = qp->rcode;
        qr->qrr.r_ourname = sp->ourname;
        qr->qrr.r_ancount = qp->ancount;
        qr->qrr.r_nscount = qp->nscount;
        qr->qrr.r_arcount = qp->arcount;
 
        /* handle the leaky bucket algorithm */
 
        if( qr->qrr.r_lbucket_count == MAX_STACK ) i = 1;
        else i = 0;
 
        for( ; i<qr->qrr.r_lbucket_count; i++ )
            if( current_time-LBUCKET_WINDOW < qr->qrr.r_stack[i] ) break;
 
        if( i > 0 ) { memcpy( qr->qrr.r_stack,
            &(qr->qrr.r_stack[i]), sizeof(qr->qrr.r_stack) - i*sizeof(u_long) );
            qr->qrr.r_lbucket_count -= i; }
 
        qr->qrr.r_stack[qr->qrr.r_lbucket_count] = current_time;
        qr->qrr.r_lbucket_count++;
 
        /* If there were no errors and some returned data then update
           the domain wide cache data */
 
        if( qp->rcode == NOERROR &&
            (qp->ancount > 0 || qp->nscount > 0 ) ) {
            ipdp = qr->qrk.ipd_link;
            ipdp->r_latest_time = current_time;
            ipdp->r_ttl = qr->qrr.r_ttl;
            }
        }
 
    }

 
 
/*------------------------------------------------------------------------
--
--  delete_qr_record() - A routine to eliminate a record from the
--  query response database.
--
--  parameters:
--      qrpt - a pointer to the structure to delete
--
--  return value:
--      none.
--
--  notes:
--      none.
--
------------------------------------------------------------------------*/
void delete_qr_record( qrpt )
QR_DATA *qrpt;
    {
    QR_DATA **tmpt;
    QR_HTABLE *qrhp, **qhp;
    IPD_DATA *ipdp;
    u_long hcode;
 
    /* Extract the record from the deletion queue. */
 
    if( qrpt->qrk.tq_flink != NULL )
        (qrpt->qrk.tq_flink)->qrk.tq_blink = qrpt->qrk.tq_blink;
    if( qrpt->qrk.tq_blink != NULL )
        (qrpt->qrk.tq_blink)->qrk.tq_flink = qrpt->qrk.tq_flink;
 
    /* Get the linked list head for this record's hash queue. */
 
    qrhp = qrpt->qrk.blink;
 
    /* Decrement the reference count on the domain wide record matching
       this query's source ip address.  Delete the record if necessary. */
 
    ipdp = qrpt->qrk.ipd_link;
    ipdp->reference_count--;
    if( ipdp->reference_count <= 0 ) delete_ipd_record( ipdp, qrhp );
 
    /* Delete the query record from the hash queue. */
 
    for( tmpt = &qrhp->qrlink; *tmpt != NULL;
        tmpt = &(*tmpt)->qrk.flink )
        if( *tmpt == qrpt ) break;
    if( *tmpt == qrpt ) *tmpt = qrpt->qrk.flink;
 
    /* If the hash list has no entries on it then delete that
       as well. */
 
    if( qrhp->qrlink == NULL ) {
        hcode = qr_hash( qrhp->q_name, qrhp->q_type );
        for( qhp=&qr_tbl[hcode]; *qhp!=NULL; qhp=&(*qhp)->hclink )
            if( *qhp == qrhp ) break;
        if( *qhp == qrhp ) *qhp = qrhp->hclink;
        total_qn_chars -= (strlen(qrhp->q_name) + 1);
        free( qrhp->q_name );
        if(mfp) fprintf( mfp, "freed qr string at %X\n", qrhp->q_name );
        free( qrhp );
        if(mfp)fprintf(mfp,"freed QR_HTABLE at %X (hcode=%d)\n", qrhp, hcode );
        total_ht_records--;
        }
 
    /* Decrement the reference count on the ip address database
       entry and delete that record if we are the last contact
       with it. */
 
    (qrpt->qrk.iprec)->reference_count--;
    if( (qrpt->qrk.iprec)->reference_count == 0 )
        delete_ip_record( qrpt->qrk.iprec );
 
    /* Free up the space it was occupying. */
 
    free( qrpt );
    if(mfp) fprintf( mfp, "freed QR_DATA at %X\n", qrpt );
    total_qr_records--;
    }

 
 
/*------------------------------------------------------------------------
--
--  qr_hash - create a simple hash code from a query name and type.
--
--  parameters:
--      qr_name - character string for query name
--      qr_type - integer for query type
--
--  return value:
--      none.
--
--  notes:
--      none.
--
------------------------------------------------------------------------*/
u_long qr_hash( qr_name, qr_type )
unsigned char *qr_name;
u_long qr_type;
    {
    int i;
    u_long hcode, tmpi;
    unsigned char *tp, t;
    unsigned char *bswp;
 
    /* Sum up the contents of the query string as if they were
       unsigned 32bit quantities. Start with the query type
       value. */
 
    for( hcode=qr_type,i=0; i<(strlen(qr_name)-3); i+=4 ) {
        memcpy( &tmpi, &qr_name[i], 4 ); /*byte align*/
        hcode += tmpi; }
    
    /* Sum up the remaining characters in the query name. */
 
    for( ; i<strlen(qr_name); i++ ) hcode += qr_name[i];
 
    /* Now do a bytes swap to magnify the importance of the low byte */
 
    bswp = (unsigned char*)&hcode;
    t = bswp[0]; bswp[0] = bswp[3]; bswp[3] = t;
    t = bswp[1]; bswp[1] = bswp[2]; bswp[2] = t;
 
    /* Now mod by the hash table size to make sure that the hash code
       stays inside the predefined table length. */
 
    hcode = hcode % QR_HASH_TABLE_SIZE;
 
    /* Return the new code to the caller. */
 
    if(mfp) fprintf( mfp, "hc=%d, typ=%d, nam=%s<\n",
        hcode, qr_type, qr_name );
    return( hcode );
    }

 
 
/*------------------------------------------------------------------------
--
--  create_ipd_record - create an IPD_DATA record on the query name hash
--  table linked list.
--
--  parameters:
--      ip_addr - remote ip address to use as source for domain name.
--      hp      - pointer to hash table list entry
--
--  return value:
--      pointer to the newly created record or NULL if out of memory.
--
--  notes:
--      none.
--
------------------------------------------------------------------------*/
IPD_DATA *create_ipd_record( ip_addr, hp )
u_long ip_addr;
QR_HTABLE *hp;
    {
    IPD_DATA *ipdp;
 
    /* Allocate a record and put the creation time in it */
 
    ipdp = (IPD_DATA *)calloc( 1, sizeof(IPD_DATA) );
    if( ipdp == NULL ) { return( NULL ); }
    if(mfp) fprintf( mfp, "allocated %X to %X for IPD_DATA\n",
        ipdp, ipdp+sizeof(IPD_DATA)-1 );
    ipdp->q_first_time = current_time;
 
    /* Put the domain address in the new record */
 
    ipdp->remote_domain = what_class( ip_addr );
 
    /* Link the new record at the top of the linked list. */
 
    ipdp->flink = hp->ipdlink;
    hp->ipdlink = ipdp;
    total_dm_records++;
 
    /* Return the record's pointer */
 
    return( ipdp );
    }

 
 
/*------------------------------------------------------------------------
--
--  find_ipd_record - find an IPD_DATA record on the query name hash
--  table linked list.
--
--  parameters:
--      ip_addr - remote ip address to use as source for domain name.
--      hp      - pointer to hash table list entry
--
--  return value:
--      pointer to the requested record or NULL if not found.
--
--  notes:
--      none.
--
------------------------------------------------------------------------*/
IPD_DATA *find_ipd_record( ip_addr, hp )
u_long ip_addr;
QR_HTABLE *hp;
    {
    IPD_DATA *ipdp;
    u_long daddr;
 
    /* Turn the ip address into a domain address */
 
    daddr = what_class( ip_addr );
 
    /* Look down the linked list of IPD structures for a matching
       address. */
 
    for( ipdp=hp->ipdlink; ipdp != NULL; ipdp=ipdp->flink )
         if( ipdp->remote_domain == daddr ) break;
 
    /* Return the resulting pointer. */
 
    return( ipdp );
    }

 
 
/*------------------------------------------------------------------------
--
--  delete_ipd_record - delete an IPD_DATA record from the query name hash
--  table linked list.
--
--  parameters:
--      ipdp    - pointer to the record to delete
--      hp      - pointer to hash table list entry
--
--  return value:
--      none.
--
--  notes:
--      none.
--
------------------------------------------------------------------------*/
int delete_ipd_record( ipdp, hp )
IPD_DATA *ipdp;
QR_HTABLE *hp;
    {
    IPD_DATA *ipdtp, **plink;
 
    /* Look down the linked list of IPD structures on the indicated
       hash table entry to find the matching pointer.  Make sure to
       remember the previous entry so that we can relink up any
       following structures. */
 
    plink = &(hp->ipdlink);
    for( ipdtp=hp->ipdlink; ipdtp != NULL; ipdtp=ipdtp->flink ) {
        if( ipdtp == ipdp ) break;
        plink = &(ipdtp->flink);
        }
 
    /* If we could not find a matching record then the database is
       messed up.  Just return and hope we can run a little longer. */
 
    if( ipdtp == NULL ) {
        fprintf( stderr, "delete_ipd_record: could not find record\n" );
        return; }
 
    /* Put the forward link of the record to be deleted in the previous
       record's forward link. */
 
    *plink = ipdtp->flink;
    total_dm_records--;
 
    /* Now recover the space used by the record. */
 
    free( ipdtp );
    if(mfp) fprintf( mfp, "freed IPD_DATA at %X\n", ipdtp );
 
    }

 
 
/*------------------------------------------------------------------------
--
--  sort_qr_list() - A routine to sort the link list of QR_DATA records
--  in descending order with the most prolific client's record first
--  on the list.
--
--  parameters:
--      hcptr - pointer to QR_HTABLE record for the name to be sorted.
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void sort_qr_list( hp )
QR_HTABLE *hp;
    {
    int bflag=1;
    QR_DATA *sp, **pp, *tptr;

    /* TOP OF OUTER LOOP, a simple bubble sort is used here because the
       number of clients for a name shouldn't be too long. */

    while( bflag )
        {

        /* Reset the sort looping flag on the hopes that this will be the
           last pass through the loop. */

        bflag = 0;

        /* Initialize the pointers that we are going to use to traverse the
           linked list of query structures.  */

        pp = &hp->qrlink;
        sp = (*pp)->qrk.flink;

        /* TOP OF INNER LOOP, loop down the chain of query data structures
           until the last pointer is hit (which will be NULL). */

        while( sp != NULL )
             {

             /* Test to see if the number of queries made by this client
                is more than the one preceding us on the list. */

             if( (*pp)->qrq.q_count < sp->qrq.q_count )
                 {

                 /* Swap and update our sort pointers */

                 tptr = sp->qrk.flink;
                 sp->qrk.flink = *pp;
                 (*pp)->qrk.flink = tptr;
                 *pp = sp;
                 sp = sp->qrk.flink;
                 bflag = 1;
                 }

             /* Move on to the next spot on the linked list. */

             pp = &(*pp)->qrk.flink;
             sp = sp->qrk.flink;
             }
        }
    }
