/*-------------------------------------------------------------------------
--
--  checker.c - The entry point for the runtime DNS pattern analysis
--  system.  This file also includes some functions to aid decoding
--  of the DNS message.
--
===========================================================================
--                           UPDATE HISTORY                              --
--                                                                       --
-- VER     Date         Explanation of Changes              Author       --
---------------------------------------------------------------------------
--  0   | 12/5/92  | Initial version                 | Steve Miller - USC   
--      |          |                                 |
-------------------------------------------------------------------------*/
 
#include <stdio.h>
#include <malloc.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <netdb.h>
#include "checker.h"
 
FILE *mfp = NULL;
static char srvname[200];
static u_long srvaddr = 0;
static u_long checker_enabled = 0;
 
extern char   output_file_name[100];
extern char   logging_file_name[100];
extern char   dig_path[100];
extern u_long qr_deletion_time;
extern u_long nocheck_authority;
extern u_long echeck_mask;
extern u_long ck_debug_flag;
extern u_long translate_address;
extern u_long disable_tq_flag;
extern u_long error_threshold;
extern u_long pec_checking;
extern u_long pec_blocked_time;
extern u_long pec_test_length;
extern u_long pec_interval;
extern u_long memory_logging;
extern u_long local_domain_checking;
extern u_long port_53_checking;
extern u_long maintenance_interval;
 
u_long get_answer_ttl();
char *get_query();
u_short q_type();
char *itoc();

 
 
/*------------------------------------------------------------------------
--
--  checker() - The entry point into the set of functions that
--  update and maintain the query/response database.
--
--  parameters:
--      ctype - flag showing query/response (query=0, response=1)
--      mptr - pointer to the message to be processed.
--      mlen - length in bytes of message
--      saddr - socket address structure.
--      proto - flag showing protocol type.      
--
--  globals:
--      none
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
u_long checker( ctype, mptr, mlen, saddr, qsp )
u_long ctype;
struct qstream *qsp;
struct sockaddr_in *saddr;
char *mptr;
int mlen;
    {
    QUERY_PACKET qp;
    SOURCE_INFO  src;
    struct timeval tm;
    HEADER *dh;
    char *ptr, *cp, *eom;
    int remain, control_code;
    static int init=0;
    u_short typr, qclass;
    char name[1024], tnam[20];
    u_long ptype;
    u_long stat;
    FILE *lfp;
    struct hostent *hptr;
 
    /* If we haven't initialized yet then do so. */
    
    if( !init ) {
        read_configuration();
        gethostname( srvname, sizeof(srvname) );
        hptr = gethostbyname( srvname );
        if( hptr > 0 ) srvaddr = *(u_long *)hptr->h_addr_list[0];
        init = 1; }
 
    /* Make sure we don't have garbage in our control structures. */
    
    memset( &qp, 0, sizeof(QUERY_PACKET) );
    memset( &src, 0, sizeof(SOURCE_INFO) );
 
    /* Set up pointers to the DNS message header and body */
 
    dh = (HEADER *)mptr;
    ptr = mptr + sizeof(HEADER);
    cp = ptr;
    eom = mptr + mlen;
    remain = mlen - sizeof(HEADER);
   
    /* Fill in the current time, and host address. */
    
    gettimeofday( &tm, (struct timezone *)0 );
    src.current_time = tm.tv_sec;
    src.our_address = srvaddr;
    
    /* Get the remote node's IP address and port number. */
    
    src.ip_address = saddr->sin_addr.s_addr;
    src.portnum = ntohs(saddr->sin_port);
    
    /* Extract the query name and type from the message, if we get an
       error here then assume the message is badly formed and just
       ignore it. */
    
    if( get_query(ptr,remain,&ptype,&qclass,name) == NULL) return;
 
    /* Check the returned name for nul */
 
    if( strcmp(name,"")==0) { strcpy(name, "NONE"); }
 
    /* Check to see if the query name is actually a control command to
       the checker program.  If it is then execute the command and return
       to the caller. */
       
    if( strncmp( name, "checker.control.command.", 24 ) == 0 ) {
        if( ctype == 0 ) {
            sscanf( &name[24], "%d", &control_code );
            switch( control_code ) {
                case 1: read_configuration(); break;
                case 2: dump_tables( ); break;
                }
            }
        return(0);
        }
 
    /* If our enable/disable flag is not set then return to the caller
       with no further actions. */
       
    if( !checker_enabled ) return;
 
    /* Move the query type and name into the structure fields. */
 
    qp.type = ctype;
    strcpy( qp.qname, name );
    qp.qtype = q_type(&ptype);
 
    /* Get the fields in the header of the DNS message packet. */    
    
    qp.id = dh->id;
    qp.aa = dh->aa;
    qp.rcode = dh->rcode;
    qp.qdcount = dh->qdcount;
    qp.ancount = dh->ancount;
    qp.nscount = dh->nscount;
    qp.arcount = dh->arcount;
 
    /* get a value to use in some of the cache and error checks. */
 
    if( dh->rcode == NOERROR )
        qp.ttl = get_answer_ttl( ptr, eom, &cp, (int)dh->ancount );
    else qp.ttl = NEGATIVE_TTL;
    
    /* Figure out what protocol we are dealing with and if the message
       is a response then parse out the nameserver fields to see if our
       own name is in there. */
 
    if( qsp == (struct qstream *)0 ) src.protocol = 'U';
    else                             src.protocol = 'T';
    if( ctype == RESPONSE_MESSAGE )
        src.ourname = get_ourname( ptr, eom, &cp, (int)dh->nscount ); 
        
    /* If the debugging flag is on then dump a copy of the control
       structures. */
       
    if( ck_debug_flag ) {
        if( (lfp=fopen(logging_file_name,"a")) != NULL ) {
            if( qp.type == 0 ) fprintf( lfp, "src: -----> " );
            else               fprintf( lfp, "src: <----- " );
            fprintf( lfp, "%s\n", ctime(&src.current_time));
            itoc( &src.ip_address, 4, tnam );
            fprintf( lfp, "src: from %s at port %d, proto=%d\n",
                tnam, (int)src.portnum, (int)src.protocol );        
            itoc( &src.our_address, 4, tnam );
            fprintf( lfp, "src: our_address = %s, ourname = %d\n",
                tnam, (int)src.ourname );
            fprintf( lfp, "qpk: type=%d, qtype=%d, id=%d, aa=%d, ttl=%d\n",
                qp.type, qp.qtype, qp.id, qp.aa, qp.ttl );
            fprintf( lfp, "qpk: rcode=%d, qd=%d, an=%d, ns=%d, ar=%d\n",
                qp.rcode, qp.qdcount, qp.ancount, qp.nscount, qp.arcount );
            fprintf( lfp, "qpk: %s\n\n", qp.qname );
            fclose( lfp );
            }
        }
 
    /* Now dive into the actual statistical checking code. */
    
    stat = handle_packet( &qp, &src );
 
    /* The status is an indication of whether or not the query name is
       under interdiction by the PEC algorithm.  Return it to the caller */
 
    return( stat );
    }

 
 
/*-----------------------------------------------------------------------
--
--  A routine to read the configuration file and set the variables
--  as indicated. 
--
-----------------------------------------------------------------------*/
int read_configuration()
    {
    FILE *fp, *lfp;
    char line[100];
    char *t1, *t2, t3[100];
    u_long tmp;
    char tmp1[100];
    static char eqstring[40] = "======================================";
    FILE *open_logging_file();
 
    /* Go find the configuration file. */
    
    if( (fp=fopen("/etc/checker.cfg","r")) == NULL ) {
        fprintf( stderr, "cannot open checker.cfg" );
        return; }
 
    /* Open the log file, use a default name if an earlier configuration
       file has not opened one already. */
 
    lfp = open_logging_file( logging_file_name );
    fprintf( lfp, "%s%s\n", eqstring, eqstring );
 
    /* Get a line and get the third token as a parameter. */
        
    while( fgets( line, 100, fp ) != NULL ) {
        if( line[0] == '#' ) continue;
        sscanf( line, "%*s %*s %s", t3 );
 
        if( strncmp( line, "debug flag", 10 ) == 0 ) {
            sscanf( t3, "%d", &ck_debug_flag );
            if(lfp) fprintf( lfp, "set debug flag to %d\n", ck_debug_flag );
            }
        
        else if( strncmp( line, "error mask", 10 ) == 0 ) {
            sscanf( t3, "%X", &echeck_mask );
            if(lfp) fprintf( lfp, "set echeck mask to %X\n", echeck_mask );
            }
        
        else if( strncmp( line, "time queue", 10 ) == 0 ) {
            if( strncmp(  "off", t3, 3 ) == 0 ) disable_tq_flag = 1; 
            else disable_tq_flag = 0;
            if(lfp) fprintf( lfp, "set disable tq flag to %d\n",
                disable_tq_flag );
            }
        
        else if( strncmp( line, "domain checking", 15 ) == 0 ) {
            if( strncmp(  "off", t3, 3 ) == 0 ) local_domain_checking = 0;
            else local_domain_checking = 1;
            if(lfp) fprintf( lfp, "domain checking flag is now %d.\n",
                local_domain_checking );
            }
        
        else if( strncmp( line, "port53 checking", 15 ) == 0 ) {
            if( strncmp(  "off", t3, 3 ) == 0 ) port_53_checking = 0;
            else port_53_checking = 1;
            if(lfp) fprintf( lfp, "port 53 checking flag is now %d.\n",
                 port_53_checking );
            }
        
        else if( strncmp( line, "checker enabled", 15 ) == 0 ) {
            if( strncmp(  "off", t3, 3 ) == 0 ) checker_enabled = 0;
            else checker_enabled = 1;
            if(lfp) fprintf( lfp, "checker enable flag is now %d.\n",
                checker_enabled );
            }
                
        else if( strncmp( line, "error threshold", 15 ) == 0 ) {
            sscanf( t3, "%d", &error_threshold );
            if(lfp) fprintf( lfp, "error threshold is %d\n", error_threshold );
            }
                
        else if( strncmp( line, "deletion time", 13 ) == 0 ) {
            sscanf( t3, "%d", &qr_deletion_time );
            if(lfp) fprintf( lfp, "qr deletion time is %d\n",
                qr_deletion_time );
            }
 
        else if( strncmp( line, "dig path", 8 ) == 0 ) {
            sscanf( t3, "%s", dig_path );
            if(lfp) fprintf( lfp, "dig path name is now %s\n",
                dig_path );
            }
 
        else if( strncmp( line, "authority check", 15 ) == 0 ) {
            if( strncmp(  "off", t3, 3 ) == 0 ) nocheck_authority = 1; 
            else nocheck_authority = 0;
            if(lfp) fprintf( lfp, "nocheck_authority = %d\n",
                nocheck_authority );
            }
 
        else if( strncmp( line, "memory logging", 14 ) == 0 ) {
            if( strncmp(  "off", t3, 3 ) == 0 ) memory_logging = 0;
            else memory_logging = 1;
            if(lfp) fprintf( lfp, "memory_logging = %d\n", memory_logging );
            }
        
        else if( strncmp( line, "translate address", 17 ) == 0 ) {
            if( strncmp(  "off", t3, 3 ) == 0 ) translate_address = 0; 
            else translate_address = 1;
            if(lfp) fprintf( lfp, "translate_address = %d\n",
                translate_address );
            }
        
        else if( strncmp( line, "pec checking", 12 ) == 0 ) {
            if( strncmp(  "off", t3, 3 ) == 0 ) {
                 if( pec_checking ) { /* shut PEC checking off */ }
                 pec_checking = 0;
                 }
            else pec_checking = 1;
            if(lfp) fprintf( lfp, "pec checking enable flag is now %d.\n",
                pec_checking );
            }
 
        else if( strncmp( line, "maintenance interval", 20 ) == 0 ) {
            sscanf( t3, "%d", &maintenance_interval );
            if(lfp) fprintf( lfp, "maintenance_interval = %d\n",
                maintenance_interval );
            }

        else if( strncmp( line, "pec block", 9 ) == 0 ) {
            sscanf( t3, "%d", &pec_blocked_time );
            if(lfp) fprintf( lfp, "pec_blocked_time = %d\n",
                pec_blocked_time );
            }
 
        else if( strncmp( line, "pec length", 10 ) == 0 ) {
            sscanf( t3, "%d", &pec_test_length );
            if(lfp) fprintf( lfp, "pec_test_length = %d\n",
                pec_test_length );
            }
 
        else if( strncmp( line, "pec interval", 12 ) == 0 ) {
            sscanf( t3, "%d", &pec_interval );
            if(lfp) fprintf( lfp, "pec_interval = %d\n",
                pec_interval );
            }
 
        else if( strncmp( line, "logging file", 12 ) == 0 ) {
            sscanf( t3, "%s", tmp1 );
            if( strcmp( tmp1, logging_file_name ) != 0 ) {
                strcpy( logging_file_name, tmp1 );
                if(lfp) {
                    fprintf( lfp, "New logging file name is %s\n",
                        logging_file_name );
                    fclose( lfp ); }
                lfp = open_logging_file( logging_file_name );
                fprintf( lfp, "%s%s\n", eqstring, eqstring );
                }
            else if(lfp) fprintf( lfp, "logging file name is %s\n",
                logging_file_name );
            }
        
        else if( strncmp( line, "output file", 11 ) == 0 ) {
            sscanf( t3, "%s", tmp1 );
            set_out_nam( tmp1 );
            if(lfp) fprintf( lfp, "output reports file name is now %s\n",
                output_file_name );
            }
        }
 
    /* close the logging and configuration files */
    
    if(lfp) {
        fprintf( lfp, "%s%s\n", eqstring, eqstring );
        fclose(lfp); }
    fclose(fp);
    }

 
 
/*-----------------------------------------------------------------------
--
--  A simple routine to open a logging file.
--
-----------------------------------------------------------------------*/
FILE *open_logging_file( logging_file_name )
char *logging_file_name;
    {
    FILE *lfp;
    struct timeval tm;
 
    /* Open the log file, use a default name if an earlier configuration
       file has not opened one already. */
        
    if( (lfp=fopen(logging_file_name,"a"))==NULL ) {
        fprintf( stderr, "cannot open %s as checker's log file",
            logging_file_name ); }
 
    return(lfp);
    }
 

 
 
/*-----------------------------------------------------------------------
--
--  A simple routine to write a logging file text entry.
--
-----------------------------------------------------------------------*/
void write_log_entry( text )
char *text;
    {
    FILE *lfp;
    struct timeval tm;
    char tstr[100];
 
    /* Open the log file. */
        
    if( (lfp=fopen(logging_file_name,"a"))==NULL ) {
        fprintf( stderr, "cannot open %s as checker's log file",
            logging_file_name ); }
            
    /* Write the current time and the desired text into the log file. */
 
    if(lfp) {
        gettimeofday( &tm, (struct timezone *)0 );
        strcpy( tstr, ctime(&tm.tv_sec) );
        tstr[24] = '\0';
        fprintf( lfp, "%s| %s\n", tstr, text );
        fclose( lfp );
        }
    }

 
 
/*-----------------------------------------------------------------------
--
--  A little routine to pull the time-to-live for the answer record
--  from a returned response.
--
-----------------------------------------------------------------------*/
u_long get_answer_ttl( ptr, eom, cp, ancount )
char *ptr, *eom, **cp;
int ancount;
    {
    u_long ttl,tttl;
    u_long _getlong();
    u_short _getshort();
    int rdlen, n;
 
    /* Assume that there is only one question record and get its fields
       out of the way. */
 
    if( (n = dn_skipname( ptr, eom )) < 0 ) return(NEGATIVE_TTL);
    *cp += n;                  /* account for name field     */
    *cp += sizeof( u_short );  /* account for type field     */
    *cp += sizeof( u_short );  /* account for class field    */
    
    /* Now loop for the number of answer records, trying to find the one
       with the least TTL.  That is the one we want to use for database
       limiting and cache leak detection. */
 
    for( ttl=0xEFFFFFFF; ancount>0; ancount-- ) {
 
        /* Read out the name field from the answer resource record.
           The name itself isn't really needed, but we have to account
           for the space. */
 
        if( (n = dn_skipname( *cp, eom )) < 0 ) break;
 
        /* Now get the ttl value and account for the rest of the 
           resource record. */
 
        *cp += n;                  /* account for name field     */
        *cp += sizeof( u_short );  /* account for type field     */
        *cp += sizeof( u_short );  /* account for class field    */
        if( (*cp)+3 > eom ) break; /* make sure we are in msg    */
        tttl = _getlong( *cp );    /* get time to live value     */
        *cp += sizeof( u_long  );  /* account for ttl field      */
        if( (*cp)+3 > eom ) break; /* make sure we are in msg    */
        rdlen = _getshort( *cp );  /* get length of rdata field  */
        *cp += sizeof( u_short );  /* account for rdlength field */
        *cp += rdlen;              /* account for rdata fields   */
 
        /* If this was a real time to live value, and is less than any
           other we have seen then save it. */
 
        if( tttl > 0 && tttl < ttl ) ttl = tttl;
        }
 
    /* If we didn't find any ttl then return the negative ttl. */
 
    if( ttl == 0xEFFFFFFF ) ttl = NEGATIVE_TTL;
 
    /* Give the ttl value back to the caller */
 
    return( ttl );
    }

 
 
/*-----------------------------------------------------------------------
--
--  A little routine to see if our own server name is in the authority
--  records section of the response.
--
-----------------------------------------------------------------------*/
int get_ourname( ptr, eom, cp, nscount )
char *ptr, *eom, **cp;
int nscount;
    {
    u_long ttl,tttl;
    u_long _getlong();
    u_short _getshort();
    int rdlen, n, found, done, type;
    char nbuf[200];
 
    /* Check to see that we are supposed to be doing this. */
    
    if( !(echeck_mask & RECURSION_C_BIT) ) return( 0 );
 
    /* Now loop for the number of authority records */
 
    for( found=0,done=0; !done && nscount>0; nscount-- ) {
 
        /* Skip the name field from the resource record.
           The name itself isn't really needed, but we have to account
           for the space. */
 
        if( (n = dn_skipname( *cp, eom )) < 0 ) break;
 
        /* Now account for the rest of the resource record. */
 
        *cp += n;                  /* account for name field     */
        if( *cp >= eom ) break;    /* make sure we don't overrun */
        type = _getshort( *cp );   /* get the record's type      */
        *cp += sizeof( u_short );  /* account for type field     */
        *cp += sizeof( u_short );  /* account for class field    */
        *cp += sizeof( u_long );   /* account for ttl field      */
        if( *cp >= eom ) break;    /* make sure we don't overrun */
        rdlen = _getshort( *cp );  /* get length of rdata field  */
        *cp += sizeof( u_short );  /* account for rdlength field */
 
        /* If this is really an NS type of record then expand the
           domain name in the rdata field to see if it matches our
           own name. */
 
        if( type == T_NS ) {
            dn_expand( ptr, eom, *cp, nbuf, sizeof(nbuf) );
            if( strncmp( nbuf, srvname, strlen(srvname) ) == 0 ) {
                done = 1; found = 1; }
            }
        *cp += rdlen;              /* account for rdata fields   */
        }
 
    /* Give the found/not-found status back to the caller */
 
    return( found );
    }

 
 
/*------------------------------------------------------------------------
--
--  get_query - a routine to extract the contents of the question type of
--  resource record that is pointed to by the input pointer.
--
--  Note that this routine assumes that this is the first name in the
--  message and does not attempt to discover or resolve compressed
--  labels.
-- 
--  anant jan 19, modify to also accept qclass pointer as an
--  argument. class value is written and thus passed back.
--
-------------------------------------------------------------------------*/
char *
get_query(ptr, remain, typep, qclass, name)
char *ptr;
int remain;
u_short *qclass;
u_short *typep;
char *name;
    {
    int i;
    int qlen, len;
    char label[256];
    int cnt=0;
 
    /* Make sure we start out clean. */
 
    *name = '\0'; *label = '\0';
 
    /* Loop through all of the labels (those pieces normally separated
       by dots in our common Domain Name structure. */
 
    for( i=0; ; i++ ){
 
        /* Get the length of the label and check that its not too long */
 
        qlen = (u_char) (*ptr);
        if( qlen > 63 ) {
          sprintf( name, "LONG COMPONENT");
          return(name); }
 
        /* Add that to the total length and check that the whole name
           is not too long. */
 
        cnt += qlen;
        if( cnt > 1024 ) {
            sprintf( name, "LONG NAME");
            return(name); }
 
        /* Update our byte pointers into the message, and break out of
           the loop if we have found the end of the name. */
 
        ptr++; remain--;
        if (qlen == 0) break;
 
        /* Copy this component of the name into our working buffer. */
 
        bcopy( ptr, label, qlen );
        label[qlen] = '\0';
        ptr        += qlen;
        remain     -= qlen;
 
        /* If this is the first component of the name then just copy
           the label into the output buffer.  If not then add the
           dot before this component and concatenate with the current
           contents of the output buffer. */
 
        if( i == 0 ){ strncpy(name, label, qlen+1); }
        else {
            strcat(name,".");
            strncat(name, label, qlen+1);
            }
        }
 
    /* Now recover the query type and class from the end of the record
       in the message. */
 
    len = sizeof(u_short);
    bcopy(ptr, typep, len);
    ptr+=len; remain -= len;
    bcopy(ptr, &qclass, len);
    ptr += len; remain -= len;
    return(name);  
    }

 
 
/*-----------------------------------------------------------------------
--
--  q_type - decode the query type field in a question record
--
-----------------------------------------------------------------------*/
u_short
q_type(typep)
u_short *typep;
    {
    u_short vtype;
  
    if ((*typep < 17) && (*typep > 0))
        vtype = *typep;
    else if ((*typep > 251) && (*typep < 256))
        vtype = *typep - 235; /* - 252 + 16 + 1 */
    else vtype = 0;
    return(vtype);
    }

 
 
/*-----------------------------------------------------------------------
--
--  itoc - returns inet address in dot format given HEX address
--
-----------------------------------------------------------------------*/
char *
itoc(str, len, rlt)
char *str;
int len;
char *rlt;
    {
    int i;
    char asc[8];
  
    rlt[0] = '\0';
    for( i=0; i < len; i++) {
        sprintf(asc, "%d.", (u_char)str[i]);
        strcat(rlt, asc); }
 
    rlt[strlen(rlt) - 1] = '\0';
    return(rlt);
    }

 
 
/*-----------------------------------------------------------------------
--
--  A simple routine to dump out the current state of the malloc
--  buffer space taken from the heap.
--
-----------------------------------------------------------------------*/
dump_mallinfo()
    {
    struct mallinfo mi;
    struct timeval tm;
    FILE *lfp;
    char lstr[100];
    
    /* Get the malloc database information structure. */
    
    mi = mallinfo();
    
    /* Format an output string and write it to the log file. */
    
    sprintf( lstr, "uordblks = %d", mi.uordblks );
    write_log_entry( lstr );
 
/*  WARNING: Some items here are SunOS specific.
    printf( "arena     = %d\n", mi.arena );
    printf( "ordblks   = %d\n", mi.ordblks );
    printf( "smblks    = %d\n", mi.smblks );
    printf( "hblks     = %d\n", mi.hblks );
    printf( "hblkhd    = %d\n", mi.hblkhd );
    printf( "usmblks   = %d\n", mi.usmblks );
    printf( "fsmblks   = %d\n", mi.fsmblks );
    printf( "uordblks  = %d\n", mi.uordblks );
    printf( "fordblks  = %d\n", mi.fordblks );
    printf( "keepcost  = %d\n", mi.keepcost );
    printf( "mxfast    = %d\n", mi.mxfast );
    printf( "nlblks    = %d\n", mi.nlblks );
    printf( "grain     = %d\n", mi.grain );
    printf( "uordbytes = %d\n", mi.uordbytes );
    printf( "allocated = %d\n", mi.allocated );
    printf( "treeoverh = %d\n", mi.treeoverhead );
*/
    }

