/*-------------------------------------------------------------------------
--
--  checker.h - A set of constants and structures used to implement the
--  query/response and server/resolver databases.
--
--  The checker databases consist of the following:
--      1) a database holding query names, response values, and
--         all necessary information to do the bug checking functions
--         of the checker program.  This database is implemented
--         using the QR_HTABLE pointing to QR_DATA (individual queries)
--         and IPD_DATA (domain wide) structures.
--      2) a database holding the ip addresses of the various name
--         servers and resolvers that have made use of the name server
--         hosting checker().  This database is built from the
--         IP_DATA type of structures.
--      3) a set of linked lists that thread through all records in the
--         QR_DATA database, ordered by the time at which they should be
--         purged from the database. When a QR_DATA record is deleted the
--         reference count in the IP_DATA record it points to is
--         decremented and that record also deleted if necessary.
--
===========================================================================
--                           UPDATE HISTORY                              --
--                                                                       --
-- VER     Date         Explanation of Changes              Author       --
---------------------------------------------------------------------------
--  0   |  7/20/92 | Initial version                 | Steve Miller - USC
--  1   |  7/24/92 | eliminated tree, added hash tbl | Steve Miller - USC
--  2   |  9/20/92 | added IPD domain wide stuff     | Steve Miller - USC
--      |          | and NO_CENTRAL_CACHE checks     |   
--  3   | 10/23/92 | added counters and FORMERR chk  | Steve Miller - USC
--  4   |  1/8/93  | added PEC code                  | Steve Miller - USC
--      |          |                                 |
-------------------------------------------------------------------------*/


/*-------------------------------------------------------------------------
--
--  Constants used in the checker programs
--
-------------------------------------------------------------------------*/

/* Classifications for incoming message types used to decide what type of
   error mode checking is to be done on the incoming data. */

#define UNASSIGNED   -1  /* used in intermediate processing in code    */
#define NOT_PROCESSED 0  /* query/response message is not handled      */
#define INITIAL_MSG   1  /* message is first one for this machine      */
#define INITIAL_QUERY 2  /* message is first one for this name/type    */
#define QUERY_UPDATE  3  /* message is a second query for known name   */
#define RESPONSE      4  /* message is a response to earlier query     */

/* Hash table array sizes for the two main databases used in the checker
   programs. */

#define QR_HASH_TABLE_SIZE 20997
#define IP_HASH_TABLE_SIZE 20997

/* Number of slots in the stack of respones that are used to determine
   that a resolver is recursively querying beyond the recomended depth. */ 

#define MAX_STACK 10   /* maximum number time entries in leaky bucket    */
#define MAX_LB_COUNT 6 /* numbr of entries in LB needed to trigger error */
#define LBUCKET_WINDOW 1000 /* sec, window on times for recursion check  */

/* Number of slots in the stack of respones that are used to determine
   that a node is ignoring the fact that no responses are coming from a
   given server. */

#define MAX_IPQ_STACK 10   /* maximum number time entries in leaky bucket    */
#define MAX_IPQ_COUNT 6 /* numbr of entries in LB needed to trigger error */
#define IPQ_LBKT_WINDOW 1000 /* sec, window on times for recursion check  */

/* Error mode flag bits returned from the rules checking code and used
   to generate the error reporting records returned to the clearing
   house. */

#define FAST_RX_BIT          0x0001 /* didn't wait for answer            */
#define RECURSION_A_BIT      0x0002 /* ignored referral & tried again    */
#define RECURSION_B_BIT      0x0004 /* too many referrals (recursions)   */
#define RECURSION_C_BIT      0x0008 /* host named itself in referral     */
#define ZERO_ANSWER_BIT      0x0010 /* ignored "no such record"          */
#define SERVER_FAILURE_A_BIT 0x0020 /* resolver didn't handle servfail   */
#define SERVER_FAILURE_B_BIT 0x0040 /* nameserver didn't handle servfail */
#define CACHE_LEAK_BIT       0x0080 /* requeried inside ttl value        */
#define NAME_ERROR_BIT       0x0100 /* ignored "no such domain name"     */
#define NO_CENTRAL_CACHE_BIT 0x0200 /* domain doesn't use central cache  */
#define FORMAT_ERROR_BIT     0x0400 /* Badly formed query from requestor */
#define ALL_ERRORS       0xFFFFFFFF /* All possible error bits           */

/* Error array data structure indices, used to access the error time and
   count slot allocated when an error is detected. */

#define FAST_RX_SLOT            0   /* didn't wait for answer            */
#define RECURSION_A_SLOT        1   /* ignored referral & tried again    */
#define RECURSION_B_SLOT        2   /* too many referrals (recursions)   */
#define RECURSION_C_SLOT        3   /* host named itself in referral     */
#define ZERO_ANSWER_SLOT        4   /* ignored "no such record"          */
#define SERVER_FAILURE_A_SLOT   5   /* resolver didn't handle servfail   */
#define SERVER_FAILURE_B_SLOT   6   /* nameserver didn't handle servfail */
#define CACHE_LEAK_SLOT         7   /* requeried inside ttl value        */
#define NAME_ERROR_SLOT         8   /* ignored "no such domain name"     */
#define NO_CENTRAL_CACHE_SLOT   9   /* domain doesn't use central cache  */
#define FORMAT_ERROR_SLOT      10   /* requestor sent a garbled query    */
#define NUMBER_OF_ERROR_TYPES  11   /* used to allocate array size       */

/* Text strings used to describe each error condition in the output report
   file.  These should all be the same defined length. */

#define ERR_STR_LENGTH               18
#define FAST_RX_STR          "Fast Retransmit   "
#define RECURSION_A_STR      "Ignored Referrals "
#define RECURSION_B_STR      "Too Many Referrals"
#define RECURSION_C_STR      "Host Named Self   "
#define ZERO_ANSWER_STR      "Ignored No Record "
#define SERVER_FAILURE_A_STR "Resolver NoSrvfail"
#define SERVER_FAILURE_B_STR "Namservr NoSrvfail"
#define CACHE_LEAK_STR       "Requery Inside TTL"
#define NAME_ERROR_STR       "Ignored No Domain "
#define NO_CENTRAL_CACHE_STR "No Central Cache  "
#define FORMAT_ERROR_STR     "Format Error      "


/* Path names used for the reports. */

#define DIGFILE "/tmp/digfile"
#define DEFAULT_OUTPUT_FILE "/usr/tmp/checker.stats"
#define DEFAULT_LOG_FILE "/usr/tmp/checker.log"
#define DEFAULT_DIG_PATH "/usr/etc"

/* States that a database record can be in, adjusted based on the most
   recent type of message received concerning the name or address
   that the record concerns. */

#define WAITING_FOR_RESPONSE 1
#define RESPONSE_RECEIVED    2

/* Timeout constants for the error mode checking. */

#define STDTIMOUT 30 /* sec, usual wait for an answer before requerying  */
#define DELTA      5 /* sec, processing  time to generate a requery      */

/* Deletion queue control constants. */

#define REPORT_TIMEOUT            300 /* sec, how long for a single report*/
#define NEGATIVE_TTL            86400 /* sec, how long to keep error info */
#define DEFAULT_SCAN_INTERVAL    3600 /* sec, how often to do cleanup     */
#define TIME_QUEUE_SLOT_COUNT      24 /* number of deletion queues        */
#define QR_DELETION_TIMEOUT     14400 /* sec for query to be cached       */
#define ERR_DUMP_SLOTS             18 /* dumps top n offenders            */
#define QR_DUMP_SLOTS              25 /* dumps top n query names          */
#define IP_DUMP_SLOTS              50 /* dumps top n client addresses     */

/* Domain name system values. */

#define QUERY_MESSAGE 0
#define RESPONSE_MESSAGE 1

/* Proactive Error Checking (PEC) (also known as Probabalistic Error
   Checking) state codes used in the hash table head for query name data */

#define PEC_NOT_TESTING     0
#define PEC_TESTING         1
#define PEC_RECENTLY_TESTED 2

/* PEC default testing values */

#define PEC_DEF_LENGTH      3600      /* seconds    */
#define PEC_DEF_BLOCK       3600      /* seconds    */
#define PEC_DEF_INTERVAL    3600      /* seconds    */
#define PEC_SAFETY_LIMIT      10      /* iterations */


/*-------------------------------------------------------------------------
--
--  A structure used to build the query rate information database
--
-------------------------------------------------------------------------*/
#define NUM_RATE_INFO_SLOTS 4500   /* about 31 days at 10 min. per slot */

typedef struct {
    u_long qcount;       /* The query count at itime.                  */
    u_long itime;        /* The time (in unix seconds) for the reading */
    }   RATE_INFO;


/*-------------------------------------------------------------------------
--
--  Interface structures used in the checker programs
--
-------------------------------------------------------------------------*/

/* A defined structure to be used by the caller of handle_packet() to
   pass in the stuff that is not in the dns message. */

typedef struct {
    u_long  ip_address;   /* IP address of query source or resp target */
    u_short portnum;      /* TCP/UDP port number of packet reception   */
    u_short protocol;     /* 'U' or "T'                                */
    u_short ourname;      /* 1 = our own name is in authority section  */
    u_long  current_time; /* NIT timestamp.                            */
    u_long  our_address;  /* The address of checker's host node        */
    }  SOURCE_INFO;

/* A temporary definition of a query packet until we get the real thing
   coded up. */

typedef struct {
    u_long  type;       /* packet type                                 */
    char   qname[100];  /* query name                                  */
    u_long  qtype;      /* query type                                  */
    u_long  id;         /* query id                                    */
    u_long  aa;         /* authority flag from response                */
    u_long  ttl;        /* time to live from response                  */
    u_long  rcode;      /* response error code                         */
    u_long  qdcount;    /* question record count from query            */
    u_long  ancount;    /* answer record count from response           */
    u_long  nscount;    /* nameserver record count from response       */
    u_long  arcount;    /* additional record count from response       */
    }  QUERY_PACKET;


/*--------------------------------------------------------------------------
--
--  Predefinitions to allow some of the structures to refer to themselves.
--
--------------------------------------------------------------------------*/

struct _qr_data;
typedef struct _qr_data QR_DATA;
struct _ip_data;
typedef struct _ip_data IP_DATA;
struct _qr_htable;
typedef struct _qr_htable QR_HTABLE;
struct _ipd_data;
typedef struct _ipd_data IPD_DATA;
struct _pec_control;
typedef struct _pec_control PEC_CONTROL;


/*--------------------------------------------------------------------------
--
--  Query name and response values database.
--
--  Six structures are defined to keep track of queries and responses:
--      1) QR_HTABLE is used in the form of an array to create a hash
--         table with the query name and type being the inputs to the
--         hash algorithm.  This table entry points to a singly linked
--         list of query data records that use an IP address as their
--         key and are made up of the following five structures.
--      2) QR_KEY holds an IP address, a forward link to the next IP
--         address that happened to query for the same name and type,
--         and a backward link to the hash table entry (used when the
--         record has to be deleted.  This structure also hold the
--         pointers that make up the deletion queue.
--      3) QR_STATE holds the current and previous "state" of the query.
--         Currently the only states are WAITING_FOR_RESPONSE, and
--         RESPONSE_RECEIVED.  It is possible that some error states
--         will be defined.
--      4) QR_QUERY_DATA holds the time and id information for a given
--         query that allows some of the error checking to be done.
--      5) QR_RESPONSE_DATA holds those fields that are retained from
--         any response generated by the original query.
--      6) QR_DATA is a collection of the previous four structures, and
--         is the actual structure pointed to by the hash table.
--
--------------------------------------------------------------------------*/


struct _qr_htable            /* QR_HTABLE - used as hash table array      */
    {                        /* into the database                         */
    u_long   pec_flag;       /* proactive error checking flag             */
    u_long     q_type;       /* query type from message                   */
    char      *q_name;       /* pointer to text query name                */
    QR_DATA   *qrlink;       /* query data records for this name          */
    IPD_DATA *ipdlink;       /* domain wide information for this query    */
    QR_HTABLE *hclink;       /* forward link for hash collision           */
    };

typedef struct               /* QR_KEY - a structure used as a key        */
    {                        /* into the database                         */
    u_long remote_ip_addr;   /* ip address of querying machine            */
    u_long last_access_time; /* recorded access time for this record.     */
    QR_DATA   *flink;        /* forward link for address collision        */
    QR_HTABLE *blink;        /* backward link for deletion purposes       */
    IP_DATA   *iprec;        /* pointer to ip address rec for cleanup     */
    QR_DATA   *tq_flink;     /* forward link for time queue               */
    QR_DATA   *tq_blink;     /* backward link for time queue              */
    IPD_DATA  *ipd_link;     /* pointer to domain wide information        */
    }  QR_KEY;

typedef struct               /* QR_STATE - a structure used to            */
    {                        /* show a query's state                      */
    u_short state;           /* current state of query                    */
    u_short previous_state;  /* previous state of query                   */
    }  QR_STATE;

typedef struct               /* QR_QUERY_DATA - holds times and ids       */
    {                        /* from the most recent query                */
    u_long  q_count;         /* count of times this query seen            */
    u_short q_port;          /* socket port number used                   */
    u_short q_protocol;      /* type of transmission protocol used        */
    u_short q_id;            /* requestor's id field from query           */
    u_short q_previous_id;   /* previous id of similar query              */
    u_long  q_first_time;    /* first time query was seen                 */
    u_long  q_latest_time;   /* most recent time query was seen           */
    u_long  q_previous_time; /* previous time query was seen              */
    }  QR_QUERY_DATA;

typedef struct               /* QR_RESPONSE_DATA - data fields from       */
    {                        /* most recent response to query             */
    u_char  r_aa;            /* authoritative answer flag from msg        */
    u_long  r_ttl;           /* time to live from answer or ns rcrd       */
    u_long  r_count;         /* number of responses seen                  */
    u_char  r_rcode;         /* response code from message                */
    u_char  r_ourname;       /* flag showing our own name in ns rec       */
    u_short r_ancount;       /* number of answer rcrds in response        */
    u_short r_nscount;       /* number of authority rcrds in resp.        */
    u_short r_arcount;       /* number of additional rcrds in resp.       */
    u_long  r_latest_time;   /* time of most recent response              */
    u_long  r_previous_time; /* time of previous response                 */
    u_long  r_lbucket_count; /* number of entries in bucket stack         */
    u_long  r_stack[MAX_STACK+1]; /* stack of response times              */
    }  QR_RESPONSE_DATA;

struct _qr_data             /* QR_DATA - top level strucuture for query   */
    {                       /* database, pointed to by hash array         */
    QR_KEY           qrk;   /* database key                               */
    QR_STATE         qrs;   /* state of query                             */
    QR_QUERY_DATA    qrq;   /* fields from most recent query message      */
    QR_RESPONSE_DATA qrr;   /* fields from most recent response message   */
    };


/*--------------------------------------------------------------------------
--
--  Error description sub-structure.
--
--  A structure defined to keep track of the errors detected for a
--  given source node, used in an array indexed by a defined constant.
--  The array is only allocated when an error is detected,
--  and is maintained as long as the anchoring IP_DATA record exists.
--
--------------------------------------------------------------------------*/

typedef struct
    {
    u_long ftime;  /* first time the error was detected.                  */
    u_long ltime;  /* latest time the error was detected.                 */
    u_long count;  /* number of detections                                */
    } EARY;

typedef struct {
    EARY ea[NUMBER_OF_ERROR_TYPES];
    } ERR_DATA;


/*--------------------------------------------------------------------------
--
--  Requesting network ip address database.
--
--  A structure defined to keep track of the addresses for
--  machines requesting name server activity:
--      1) IP_DATA is a collection of: key, state, and times needed to
--         make decisions on the SERVER_FAILURE_B error mode.
--
--------------------------------------------------------------------------*/

struct _ip_data              /* IP_DATA - top level structure for ip-addr */
    {                        /* database, pointed to by hash array        */
    u_long  remote_ip_addr;  /* be used as key into the database          */
    u_short state;           /* current state                             */
    u_short previous_state;  /* previous state                            */
    u_short srcflags;        /* 0x1 = port 53, 0x2 = non port 53          */
    u_short q_lbucket_count; /* number of entries in query stack          */
    u_long  q_stack[MAX_IPQ_STACK+1]; /* stack of query times             */
    u_long  q_first_time;    /* first time a query was seen               */
    u_long  q_latest_time;   /* most recent time a query was seen         */
    u_long  q_previous_time; /* time of previous query                    */
    u_long  reference_count; /* number of different query records active  */
    u_long  tq;              /* Total number of queries by this address   */
    u_long  tr;              /* Total number of responses to this addr.   */
    u_long  neq;             /* Total number of non error queries         */
    u_long  ner;             /* Total number of non error responses       */
    u_long  sav_tq;          /* Saved number of queries during pec tests  */
    u_long  sav_tr;          /* Saved number of responses                 */
    u_long  sav_neq;         /* Saved number of non error queries         */
    u_long  sav_ner;         /* Saved number of non error responses       */
    char    *resolved_name;  /* DNS name for remote IP address            */
    ERR_DATA *err_data;      /* pointer to error data, if necessary       */
    ERR_DATA *pec_save;      /* pointer to error data recorded before pec */
    IP_DATA  *flink;         /* forward link for hash collision           */
    };


/*--------------------------------------------------------------------------
--
--  Requesting network domain wide database.
--
--  A structure defined to keep track of the time and ttl data on a
--  domain wide basis:             4
--
--------------------------------------------------------------------------*/

struct _ipd_data             /* IPD_DATA - top level structure for domain */
    {                        /* wide info per query name                  */
    u_long  remote_domain;   /* be used as key into the linked list       */
    u_long  q_first_time;    /* first time a query was seen               */
    u_long  r_latest_time;   /* most recent time a response was seen      */
    u_long  r_ttl;           /* ttl in previous response                  */
    u_long  reference_count; /* number of different query records active  */
    IPD_DATA *flink;         /* forward link in linked list               */
    };


/*-------------------------------------------------------------------------
--
--  Proactive Error Checking (PEC) control structures.
--
-------------------------------------------------------------------------*/

struct _pec_control
    {
    QR_HTABLE   *pec_victim;  /* pointer to the query name being blocked */
    u_long      start_time;   /* start time of test                      */
    u_long      end_time;     /* projected end time of test              */
    PEC_CONTROL *next;        /* link for next test control structure    */
    };


/*-------------------------------------------------------------------------
--
--  Routines used in the checker programs
--
-------------------------------------------------------------------------*/

u_long     handle_packet();
u_long     checker();
void       maintain_db();
void       update_db();
void       check_rules();
void       scan_time_queue();
void       delete_record();
void       dump_stats();
void       dump_errors();
void       dump_ipdata();
void       dump_qrdata();
void       dump_tq();
void       dump_tables();
void       dump_err_record();
void       dump_err_header();
void       write_log_entry();
FILE      *open_logging_file();
u_long     calc_db_size();
u_long     check_report_done();
u_long     what_class();
char      *decode_addr();
void       set_out_nam();

void       update_rate_info();
float      calc_rate();
u_long     find_slot_and_interp();
void       create_rate_chart();

IP_DATA   *get_ip_record();
IP_DATA   *create_ip_record();
void       update_ip_record();
void       delete_ip_record();

QR_DATA   *get_qr_record();
QR_DATA   *create_qr_record();
void       update_qr_record();
void       delete_qr_record();
void       sort_qr_list();

void       scan_pec_state();
void       pec_terminate();
void       pec_report();
void       pec_start();
void       pec_copy_stats();
QR_HTABLE *pec_select();
