/*-------------------------------------------------------------------------
--
--  check_rules.c - functions to examine database entries to see if they
--  fail any of the error checking rules.
--
--  notes:
--
===========================================================================
--                           UPDATE HISTORY                              --
--                                                                       --
-- VER     Date         Explanation of Changes              Author       --
---------------------------------------------------------------------------
--  0   |  7/20/92 | Initial version                 | Steve Miller - USC
--  1   |  9/20/92 | Added NO_CENTRAL_CACHE check    | Steve Miller - USC 
--      |          |                                 |
-------------------------------------------------------------------------*/
#include <stdio.h>
#include <malloc.h>
#include <sys/types.h>
#include <arpa/nameser.h>
#include "checker.h"
 
extern FILE *mfp;
extern u_long current_time;
extern u_long echeck_mask;
extern u_long nocheck_authority;
extern u_long ck_debug_flag;
extern u_long total_er_records;
 
void run_query_checks();
void run_response_checks();
void update_err_record();
 
 
 
 
/*------------------------------------------------------------------------
--
--  check_rules() - A routine to check the latest entry in a checker
--  database for timing or logic problems.
--
--  parameters:
--      qrd   - input pointer to the query/response record to be checked.
--      ipd   - input pointer to the server address record to be checked.
--      type  - flag indicating query or a response to a query
--      eflag - output integer bit pattern showing any errors detected
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void check_rules( qrd, ipd, type, eflag )
QR_DATA *qrd;
IP_DATA *ipd;
u_long type;
u_long *eflag;
    {
 
    /* If the type of message was a repeated query of some sort then run
       checks for FAST_RX, RECURSION_A, ZERO_ANSWER, SERVER_FAILURE_A,
       SERVER_FAILURE_B, CACHE_LEAK, NAME_ERROR, or NO_CENTRAL_CACHE. */
 
    if( type == QUERY_UPDATE ) run_query_checks( qrd, ipd, eflag );
 
    /* Otherwise the message must have been a response, run checks for the
       RECURSION_B and RECURSION_C type of problems. */
    
    else if( type == RESPONSE ) run_response_checks( qrd, ipd, eflag );
 
    }
 
 
 
 
/*------------------------------------------------------------------------
--
--  run_query_checks() - A routine to check an entry in the query
--  database for timing or logic problems.
--
--  parameters:
--      qrd   - input pointer to the query/response record to be checked.
--      ipd   - input pointer to the server address record to be checked.
--      eflag - output integer bit pattern showing any errors detected
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void run_query_checks( qrd, ipd, eflag )
QR_DATA *qrd;
IP_DATA *ipd;
u_long *eflag;
    {
    int tsf=0;
 
/*------------------------------------------------------------------------
--  SERVER_FAILURE_A
--  Check for resolver's incorrect response to a server failure
--
--  The logic here is:
--      a) did we see no response to an identical query from the same 
--             remote node?
--      b) did this query originate from a resolver (not another name
--             server)?
--      c) have there been too many queries from this ip address without 
--             an intervening response. (1000 second window, threshold = 6)
--
------------------------------------------------------------------------*/
 
    if( (echeck_mask & SERVER_FAILURE_A_BIT) != 0 &&
        qrd->qrs.previous_state == WAITING_FOR_RESPONSE &&
        qrd->qrq.q_port != 53 &&
        ipd->q_lbucket_count > MAX_IPQ_COUNT ) {
        update_err_record( ipd, (u_long)SERVER_FAILURE_A_SLOT );
        *eflag += SERVER_FAILURE_A_BIT; 
        tsf = 1; }
 
 
/*------------------------------------------------------------------------
--  SERVER_FAILURE_B
--  Check for name server incorrect response to a server failure
--
--  The logic here is:
--      a) have we not received a response to the latest query of any kind 
--             from the same remote nameserver?
--      b) did this query originate from another name server?
--      c) have there been too many queries from this ip address without 
--             an intervening response. (1000 second window, threshold = 6)
--
------------------------------------------------------------------------*/
 
    if( (echeck_mask & SERVER_FAILURE_B_BIT) != 0 &&
        ipd->previous_state == WAITING_FOR_RESPONSE &&
        qrd->qrq.q_port == 53 &&
        ipd->q_lbucket_count > MAX_IPQ_COUNT ) {
        update_err_record( ipd, (u_long)SERVER_FAILURE_B_SLOT );
        *eflag += SERVER_FAILURE_B_BIT; 
        tsf = 1; }
 
 
/*------------------------------------------------------------------------
--  FAST_RX
--  Check for too quick of a requery (didn't wait for answer)
--
--  The logic here is:
--      a) did we NOT detect any server failure errors
--      b) did we see no response to an identical query from the same 
--             remote node?
--      c) did this query happen within about 30 seconds of the previous
--             one?
--
------------------------------------------------------------------------*/
 
    if( (echeck_mask & FAST_RX_BIT) != 0 && !tsf &&
        qrd->qrs.previous_state == WAITING_FOR_RESPONSE &&
        qrd->qrq.q_latest_time < qrd->qrq.q_previous_time+STDTIMOUT ) {
        update_err_record( ipd, (u_long)FAST_RX_SLOT );
        *eflag += FAST_RX_BIT; }
 
 
/*------------------------------------------------------------------------
--  RECURSION_A
--  Check for code that ignores referral and tries our machine again
--
--  The logic here is:
--      a) have we received a response to the latest identical query 
--             from the same remote node?
--      b) was the previous response error free?
--      c) did the previous response have no answer records
--      d) did the previous response have some referral records
--      e) did this query happen within about 35 seconds of the
--             previous identical query from the same remote node?
--
------------------------------------------------------------------------*/
 
    if( (echeck_mask & RECURSION_A_BIT) != 0 &&
        qrd->qrs.previous_state == RESPONSE_RECEIVED &&
        qrd->qrr.r_rcode == NOERROR &&
        qrd->qrr.r_ancount == 0 &&
        qrd->qrr.r_nscount >  0 &&
        qrd->qrq.q_latest_time < qrd->qrq.q_previous_time+STDTIMOUT+DELTA ){
        update_err_record( ipd, (u_long)RECURSION_A_SLOT );
        *eflag += RECURSION_A_BIT; }
 
 
/*------------------------------------------------------------------------
--  ZERO_ANSWER
--  Check for code that ignores authoritative "no such record"
--
--  The logic here is:
--      a) have we received a response to the latest identical query 
--             from the same remote node?
--      b) was the previous response error free?
--      c) did the previous response have no answer records
--      d) did the previous response have no referral records
--      e) did the previous response have no extra data records
--      f) did this query happen within 86400 seconds of the
--             previous identical query from the same remote node?
--
------------------------------------------------------------------------*/
 
    if( (echeck_mask & ZERO_ANSWER_BIT) != 0 &&
        qrd->qrs.previous_state == RESPONSE_RECEIVED &&
        qrd->qrr.r_rcode == NOERROR &&
        qrd->qrr.r_ancount == 0 &&
        qrd->qrr.r_nscount == 0 &&
        qrd->qrr.r_arcount == 0 &&
        qrd->qrq.q_latest_time < qrd->qrq.q_previous_time+NEGATIVE_TTL ) {
        update_err_record( ipd, (u_long)ZERO_ANSWER_SLOT );
        *eflag += ZERO_ANSWER_BIT;  }
 
 
/*------------------------------------------------------------------------
--  CACHE_LEAK
--  Check for resolver/server having too short of a cache time
--
--  The logic here is:
--      a) have we received a response to the latest identical query 
--             from the same remote node?
--      b) was the previous response error free?
--      c) did the previous response have answer or referral records
--      d) did this query happen within the ttl of the previous response
--
------------------------------------------------------------------------*/
 
    if( (echeck_mask & CACHE_LEAK_BIT) != 0 &&
        qrd->qrs.previous_state == RESPONSE_RECEIVED &&
        qrd->qrr.r_rcode == NOERROR &&
        (qrd->qrr.r_ancount > 0 || qrd->qrr.r_nscount > 0) &&
        qrd->qrq.q_latest_time < qrd->qrr.r_latest_time+qrd->qrr.r_ttl ) {
        update_err_record( ipd, (u_long)CACHE_LEAK_SLOT );
        *eflag += CACHE_LEAK_BIT; }
 
 
/*------------------------------------------------------------------------
--  NAME_ERROR
--  Check for code that ignores authoritative "no such domain name"
--
--  The logic here is:
--      a) have we ever received a response to an identical query from
--              the same remote node?
--      b) did the response carry an NXDOMAIN error?
--      c) did the response carry authority (Can be flagged off)
--      d) did this query happen within a negative caching time window.
--
------------------------------------------------------------------------*/
 
    if( (echeck_mask & NAME_ERROR_BIT) != 0 &&
        qrd->qrr.r_latest_time != 0  &&
        qrd->qrr.r_rcode == NXDOMAIN &&
        (qrd->qrr.r_aa == 1 || nocheck_authority) &&
        qrd->qrq.q_latest_time < qrd->qrq.q_previous_time+NEGATIVE_TTL ) {
        update_err_record( ipd, (u_long)NAME_ERROR_SLOT );
        *eflag += NAME_ERROR_BIT; }
 
 
/*------------------------------------------------------------------------
--  NO_CENTRAL_CACHE
--  Check to see if the requestor is using a central cache
--
--  The logic here is:
--      a) have we ever received a response to an identical query from
--              any node in the same network?
--      b) did this query happen within the ttl for the most recent
--             response.
--      c) are there more than one nodes querying for the same query
--
------------------------------------------------------------------------*/
 
    if( (echeck_mask & NO_CENTRAL_CACHE_BIT) != 0 &&
        qrd->qrk.ipd_link->r_latest_time != 0 &&
        current_time < (qrd->qrk.ipd_link->r_latest_time +
                        qrd->qrk.ipd_link->r_ttl) &&
        qrd->qrk.ipd_link->reference_count > 1 ) {
        update_err_record( ipd, (u_long)NO_CENTRAL_CACHE_SLOT );
        *eflag += NO_CENTRAL_CACHE_BIT; }
 
    }
 
 
 
 
/*------------------------------------------------------------------------
--
--  run_response_checks() - A routine to check the error reporting rules
--  for responses to queries.
--
--  parameters:
--      qrd   - input pointer to the query/response record to be checked.
--      ipd   - input pointer to the server address record to be checked.
--      eflag - output integer bit pattern showing any errors detected
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void run_response_checks( qrd, ipd, eflag )
QR_DATA *qrd;
IP_DATA *ipd;
u_long *eflag;
    {
 
 
/*------------------------------------------------------------------------
--  RECURSION_B
--  Check for code that allows too many referrals
--
--  The logic here is:
--      a) is this response error free?
--      b) are there too many responses in the most recent 1000 seconds?
--
------------------------------------------------------------------------*/
 
    if( (echeck_mask & RECURSION_B_BIT) != 0 &&
        qrd->qrr.r_rcode == NOERROR &&
        qrd->qrr.r_lbucket_count > MAX_LB_COUNT ) {
        update_err_record( ipd, (u_long)RECURSION_B_SLOT );
        *eflag += RECURSION_B_BIT; }
 
 
/*------------------------------------------------------------------------
--  RECURSION_C
--  Check for our own host putting our own name in a referral
--
--  The logic here is:
--      a) is this response error free?
--      b) are there no answer records?
--      c) was our own node name in the name server referral records?
--
--  THIS IS TURNED OFF UNLESS THE PDUMP COMMAND LINE HAS A NODE NAME
--  SPECIFIED BEHIND THE -ma SWITCH.
--
------------------------------------------------------------------------*/
 
    if( (echeck_mask & RECURSION_C_BIT) != 0 &&
        qrd->qrr.r_rcode == NOERROR &&
        qrd->qrr.r_ancount == 0 &&
        qrd->qrr.r_ourname == 1 ) {
        update_err_record( ipd, (u_long)RECURSION_C_SLOT );
        *eflag += RECURSION_C_BIT; }
 
 
/*------------------------------------------------------------------------
--  FORMAT_ERROR
--  Check for checker's host returning a FORMERR (badly formed query) to
--  the requesting node.
--
--  The logic here is:
--      a) does this response have an error code of FORMERR?
--
------------------------------------------------------------------------*/
 
    if( (echeck_mask & FORMAT_ERROR_BIT) != 0 &&
        qrd->qrr.r_rcode == FORMERR ) {
        update_err_record( ipd, (u_long)FORMAT_ERROR_SLOT );
        *eflag += FORMAT_ERROR_BIT; }
    }
 
 
 
 
/*------------------------------------------------------------------------
--
--  update_err_record() - A routine to allocate an error record if needed
--  and to update the time and count fields.
--
--  parameters:
--      ipd   - input pointer to the server address record to be checked.
--      eslot - array index for the detected error.
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void update_err_record( ipd, eslot )
IP_DATA *ipd;
u_long eslot;
    {
    ERR_DATA *ep;
 
    /* Allocate an error data array if one doesn't exist, don't bother if
       we can't get the space. */
 
    if( ipd->err_data == NULL ) {
        ep = (ERR_DATA *)calloc( 1, sizeof(ERR_DATA) );
        if( ep == NULL ) { return; }
        if(mfp) fprintf( mfp, "allocated %X to %X for ERR_DATA\n",
            ep, ep+sizeof(ERR_DATA)-1 );
        total_er_records++;
        ipd->err_data = ep;
        }
    else ep = ipd->err_data;
 
    /* Update the time and count fields. */
 
    if( ep->ea[eslot].ftime == 0 )
        ep->ea[eslot].ftime = current_time;
    ep->ea[eslot].ltime = current_time;
    ep->ea[eslot].count++;
    }
 

