/*-------------------------------------------------------------------------
--
--  hdl_pkt.c - A set of routines to update the query/response
--  database and then detect any errors in the usage pattern by the
--  requesting server or resolver.
--
--  notes:
--      This version of the code does not support query/response pairs
--      having qdcount greater than 1 (indicating more than one query
--      per message).
--
--
===========================================================================
--                           UPDATE HISTORY                              --
--                                                                       --
-- VER     Date         Explanation of Changes              Author       --
---------------------------------------------------------------------------
--  0   |  7/18/92 | Initial version                 | Steve Miller - USC   
--  1   |  7/24/92 | changed to hash table           | Steve Miller - USC
--  2   |  9/20/92 | added error check mask controls | Steve Miller - USC
--  3   | 11/30/92 | moved reports into reports.c    | Steve Miller - USC
--  4   |  1/8/93  | added PEC code                  | Steve Miller - USC
--      |          |                                 |
-------------------------------------------------------------------------*/
 
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include "checker.h"
 
int failflag   = 0;
int memfault   = 0;
 
QR_DATA  timeq[TIME_QUEUE_SLOT_COUNT];
IP_DATA   *ip_tbl[IP_HASH_TABLE_SIZE];
QR_HTABLE *qr_tbl[QR_HASH_TABLE_SIZE];
 
u_long total_ht_records = 0;  /* QR_HTABLE   */
u_long total_qr_records = 0;  /* QR_DATA     */
u_long total_ip_records = 0;  /* IP_DATA     */
u_long total_dm_records = 0;  /* IPD_DATA    */
u_long total_er_records = 0;  /* ERR_DATA    */
u_long total_pc_records = 0;  /* PEC_CONTROL */
u_long total_qn_chars   = 0;  /* query name string storage */
u_long total_packets    = 0;  /* total packets for chosen server */
u_long total_queries    = 0;  /* total queries actually processed */
u_long total_responses  = 0;  /* total responses actually processed */
 
u_long current_time;
u_long qr_deletion_time = QR_DELETION_TIMEOUT;
u_long nocheck_authority = 0;
u_long echeck_mask = ALL_ERRORS;
u_long ck_debug_flag = 0;
u_long translate_address = 0;
u_long disable_tq_flag = 0;
u_long error_threshold = 1;
u_long pec_test_length = PEC_DEF_LENGTH;
u_long pec_blocked_time = PEC_DEF_BLOCK;
u_long pec_interval = PEC_DEF_INTERVAL;
u_long pec_checking = 0;
u_long memory_logging = 0;
u_long active_report = 0;
u_long report_pid = 0;
u_long local_domain_checking = 0;
u_long port_53_checking = 0;
u_long maintenance_interval = DEFAULT_SCAN_INTERVAL;

char logging_file_name[100] = DEFAULT_LOG_FILE;
char output_file_name[100]  = DEFAULT_OUTPUT_FILE;
char dig_path[100]          = DEFAULT_DIG_PATH;
FILE *ofp = NULL;
RATE_INFO rate_info[NUM_RATE_INFO_SLOTS];
int rinext = 0;
int riwrap = 0;
 
char erstr[NUMBER_OF_ERROR_TYPES][ERR_STR_LENGTH+1] = {
    FAST_RX_STR, RECURSION_A_STR, RECURSION_B_STR, RECURSION_C_STR,
    ZERO_ANSWER_STR, SERVER_FAILURE_A_STR, SERVER_FAILURE_B_STR,
    CACHE_LEAK_STR, NAME_ERROR_STR, NO_CENTRAL_CACHE_STR, FORMAT_ERROR_STR };

 
 
/*------------------------------------------------------------------------
--
--  handle_packet() - The entry point into the set of functions that
--  update and maintain the query/response database.
--
--  parameters:
--      pp - pointer to the query message to be processed.
--     src - pointer to a structure having ip address, protocol, & port
--
--  globals:
--      The databases are used and updated
--
--  return value:
--      FOR QUERY PACKETS: always 0
--      FOR RESPONSE PACKETS: the PEC testing flag telling whether or
--          not the query name in the packet is under interdiction.
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
u_long handle_packet( pp, src )
QUERY_PACKET *pp;
SOURCE_INFO  *src;
    {
    u_long type=0;     /* type of data in query_packet                   */
    u_long eflag=0;    /* bit mapped error indication word               */
    IP_DATA *ip_ptr;  /* ptr to record in db of nameserver ip addresses */
    QR_DATA *qr_ptr;  /* ptr to record in db of query/response data     */
 
    /* If our failure indication flag is set then return immediately and
       let the nameserver attempt to continue. */
 
    if( failflag ) return(0);
 
    /* Get the current time, needed everywhere, do it just once. */
 
    current_time = src->current_time;
 
    /* Initialize the hash tables if this is the first time in, scan the
       time queue if this is anything but the first time through.  */
 
    maintain_db( );
 
    /* If any of the maintenance functions failed then return to the
       nameserver with checker out of the loop. */
 
    if( failflag ) return(0);
 
    /* Add the new data to the database, this may mean creating a new
       entry, or just move fields from the packet into an existing record */
 
    update_db( pp, src, &qr_ptr, &ip_ptr, &type );
 
    /* If this was a update to an earlier query, or a response to a query
       then go through and check the error reporting rules. */
 
    if( type == QUERY_UPDATE || type == RESPONSE )
        check_rules( qr_ptr, ip_ptr, type, &eflag );
 
    /* Update the query, response, and error counts. */
 
    if( type == RESPONSE ) { ip_ptr->tr++;
        if( !eflag ) ip_ptr->ner++; }
    else if( type != NOT_PROCESSED ) { ip_ptr->tq++;
        if( !eflag ) ip_ptr->neq++; }
 
    /* Return a 0 or the PEC flag depending on the type of query. */
 
    if( type == RESPONSE && pec_checking ) return(qr_ptr->qrk.blink->pec_flag);
    else return( 0 );
    }

 
 
/*------------------------------------------------------------------------
--
--  maintain_db() - A routine to create and initialize the hash tables
--  for the checker databases when first called, and to scan the time
--  queue when necessary during normal operation.
--
--  parameters:
--      none
--
--  globals:
--      The time queue, address database, and query database are either
--      initialized or modified.
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void maintain_db( )
    {
    static int init=0;
    static u_long next_scan_time;
 
    /* Initialize things if we have not done that yet */
 
    if( !init ) {
        init = 1;
 
        /* Clear out the hash tables (probably doesn't need to be done,
           but some compilers are weird) */
 
        memset( qr_tbl, 0, sizeof(qr_tbl) );
        memset( ip_tbl, 0, sizeof(ip_tbl) );
        memset( timeq,  0, sizeof(timeq)  );
 
        /* Initialize the time queue scan time using the current time
           and the defined constant value. */
 
        next_scan_time = current_time + maintenance_interval;
 
        /* Initialize the random number generator that is used to
           choose query names for PEC checking. */
 
        srandom( getpid() );
        }
 
 
    /* Look to see if there is a report being produced by a child process
       and then check to see if the child has exited. */
 
    if( active_report ) {
        if( check_report_done() ) {
            active_report = 0;
            write_log_entry( "report child has exited" ); }
        }
 
 
    /* Look to see if it is time to scan the time queue for entries to
       be deleted.  Also examines the state of the proactive error
       checking system to see if something has to be done. */
 
    if( current_time > next_scan_time ) {
        next_scan_time = current_time + maintenance_interval;
        update_rate_info( );
        if( memory_logging )   dump_mallinfo();
        if( !disable_tq_flag ) scan_time_queue();
        if( pec_checking )     scan_pec_state();
        if( active_report ) if( !check_report_done() )
            write_log_entry( "report has not exited yet" );
        }
    }

 
 
/*------------------------------------------------------------------------
--
--  update_db() - A routine to add a query message's information to
--  the query/response and ip address databases.
--
--  parameters:
--      pp - pointer to the DNS query message to be processed.
--      src - host derived information (ip address & protocol/port)
--      qr_ptr - Query/response database record pointer that is returned
--               to the caller.  The record is created if necessary, and
--               is updated to reflect the information in the incoming
--               message.
--      qr_ptr - Query/response database record pointer that is returned
--               to the caller.  The record is created if necessary, and
--               is updated to reflect the information in the incoming
--               message.
--      type - Type of query returned to caller:
--             NOT_PROCESSED, INITIAL_QUERY, QUERY_UPDATE, or RESPONSE
--
--  globals:
--      The time queue, address database, and query database are either
--      initialized or modified.
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void update_db( pp, src, qrptr, ipptr, type )
QUERY_PACKET *pp;
SOURCE_INFO *src;
QR_DATA **qrptr;
IP_DATA **ipptr;
u_long *type;
    {
    u_long tadr;
 
     /* Keep track of the total number of packets that we have seen. */
 
    total_packets++;
 
    /* If there is more than one question in this incoming query
       (or response) then ignore it for the moment. */
 
    if( pp->qdcount > 1 ) {
        *type = NOT_PROCESSED;
        return; }
 
    /* Drop this packet if it is from inside our domain */
 
    if( local_domain_checking && 
        (IN_CLASSC_NET&src->our_address) == (IN_CLASSC_NET&src->ip_address) ){
        /* what_class(src->our_address) == what_class(src->ip_address) ) */
        *type = NOT_PROCESSED;
        return; }
 
    /* Drop this packet if is not from another name server. */
 
    if( port_53_checking && src->portnum != 53 ) {
        *type = NOT_PROCESSED;
        return; }
 
    /* If this is a debugging run then dump out some information */
 
    if( ck_debug_flag ) {
        if( pp->type == QUERY_MESSAGE )
            printf( "::--> %010d %s(%d) %s\n", current_time,
                pp->qname, pp->qtype, decode_addr(tadr)  );
        else printf( "::<-- %010d %s(%d) %s %d (%d|%d|%d|%d)\n", current_time,
                pp->qname, pp->qtype, decode_addr(tadr), pp->rcode,
                pp->qdcount, pp->ancount, pp->nscount, pp->arcount  );
        }
 
    /* If the message type is a response then initialize our type
       flag, otherwise leave off setting it for a while. */
 
    if( pp->type == RESPONSE_MESSAGE ) *type = RESPONSE;
    else                               *type = UNASSIGNED;
    tadr = src->ip_address;
 
    /* Go find the IP record that corresponds to the address in the
       message. */
 
    *ipptr = get_ip_record( src->ip_address );
 
    /* If we could not find a record we must make sure this is a query,
       and then create one if it is a query. */
 
    if( *ipptr == NULL ) {
        if( *type == RESPONSE ) {
            *type = NOT_PROCESSED;
            return; }
        *ipptr = create_ip_record( src->ip_address );
        if( *ipptr == NULL ) {
            memfault++;
            *type = NOT_PROCESSED;
            return; }
        *type = INITIAL_MSG;
        *qrptr = NULL;
        }
 
    /* If we did find an ip address record then go try and find a query
       name record.  (We don't need to try to find a query name record
       if we have no record of the machine having called us before.) */
 
    else { *qrptr = get_qr_record( pp->qname, pp->qtype, *ipptr ); }

    /* Mark the ip record source flags according to the port number of
       the originating request. */

    if( src->portnum == 53 ) (*ipptr)->srcflags |= 1;
    else (*ipptr)->srcflags |= 2;
 
    /* If we could not find a query record for this name and type then
       create one if the message is not a response message.  If it is
       a response then there is not much we can do with it. */
 
    if( *qrptr == NULL ) {
        if( *type == RESPONSE ) {
            *type = NOT_PROCESSED;
            return; }
        *qrptr = create_qr_record( pp->qname, pp->qtype, *ipptr );
        if( *qrptr == NULL ) {
            memfault++;
            *type = NOT_PROCESSED;
            return; }
        if( *type == UNASSIGNED ) *type = INITIAL_QUERY;
        }
 
    /* If the query name record did exist in the database then call this
       action an update, which will trigger rules checking later on. */
 
    else if( *type == UNASSIGNED ) *type = QUERY_UPDATE;
 
    /* Now that we have a pointer to the query name record update its
       fields from the message. */
 
    update_qr_record( *qrptr, pp, src );
 
    /* Now that we have updated the query/response database we can update
       the ip address record properly. */
 
    update_ip_record( *ipptr, *type );
    }

 
 
/*------------------------------------------------------------------------
--
--  scan_time_queue - a routine to look though the entries at the front
--  of the time queue to see if there are any database records that can
--  be discarded.
--
--  parameters:
--      none
--
--  globals:
--      The time queue is modified.
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void scan_time_queue()
    {
    QR_DATA *tptr, *tmptr;
    static u_long slot = 0;
 
    /* If any of our deletions cause error records to be dumped then we
       should have a header for them. */
 
    dump_err_header( ofp );
 
    /* For everything pointed to by the head of the deletion queue
       slot that is currently under examination delete any records that
       have not been accessed within the configured cache time. */
 
    for( tptr=timeq[slot].qrk.tq_flink; tptr != NULL; tptr=tmptr ) {
        tmptr = tptr->qrk.tq_flink;
        if( current_time > (tptr->qrk.last_access_time+qr_deletion_time) )
            delete_qr_record( tptr );
        }
 
    /* Increment the slot to scan next time we do this, make sure that
       it stays within range. */
 
    slot = (slot+1) % TIME_QUEUE_SLOT_COUNT;
    }

 
 
/*------------------------------------------------------------------------
--
--  what_class - returns the class of the network (A,B,C) and depending
--  on the class filters out meaningless bytes
--
--  Class A net addresses are stored in only the first of the four bytes
--  Class B net addresses                        first two bytes
--  Class C net addresses                        first three bytes
--  Class D (?)                                  all four bytes
--
--  parameters:
--      binary internet address
--
--  globals:
--      The query/response, ip address, and time queue tables are read.
--
--  return value:
--      Masked address
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
u_long what_class(in_addr)
u_long in_addr;
     {
     if (IN_CLASSA(in_addr))
         return(in_addr & IN_CLASSA_NET);/*only first byte needed*/
     else if (IN_CLASSB(in_addr))      /* address is 01xxxxx--- */
         return(in_addr & IN_CLASSB_NET); /*only first 2 bytes needed*/
     else if (IN_CLASSC(in_addr))     /* address is 110xxxxx--- */
         return(in_addr & IN_CLASSC_NET);/*only first 3 bytes needed*/
     else if (IN_CLASSD(in_addr)) {   /* address is 1110xxxx--- */
         /* hmm: what's this? */
         return(in_addr);  /* all bytes are meaningful */
         }         /* check include file <netinet/in.h> */
     }

 
/*------------------------------------------------------------------------
--
--  set_out_nam() - a simple routine to save and open the output
--  reporting file.
--
--  parameters:
--      oname - output file name
--
--  globals:
--      sets the output file pointer.
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void set_out_nam( oname )
char *oname;
    {
 
    /* If there is an existing file then close it */
 
    if( ofp != NULL ) {
        fclose( ofp );
        ofp = NULL; }
 
    /* Save the name */
 
    strcpy( output_file_name, oname ); 
 
    /* Try to open it. */
 
    if( (ofp=fopen( output_file_name, "a" )) == NULL ) {
        fprintf( stderr, "could not open %s as output report file\n",
            output_file_name );
        return;
        }
 
    /* write an error report header. */
 
    dump_err_header( ofp );
    }

 
/*------------------------------------------------------------------------
--
--  update_rate_info() - Write the current time and query count into the
--  rate information database.
--
--  parameters:
--      none
--
--  globals:
--      rate_info database and index value
--      current time and current total query count
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void update_rate_info()
    {
    int nrp;
    char tstr[100];
    static int init=0;

    /* For the first entry put a time in the past with a 0 count value. */

    if( !init ) {
        init = 1;
        rinext = 1;
        rate_info[0].qcount = 0;
        rate_info[0].itime = current_time - maintenance_interval;
        sprintf( tstr, "idx = 0    itime = %d, qcount = %d", 
             rate_info[rinext].itime, rate_info[rinext].qcount );
        write_log_entry( tstr );
        }
 
    /* Push the time in the record to be overwritten into the next further
       record, and then zero the next further record's query count.  This has
       the effect of leaving rinext (at the end of this routine) pointing to a
       record that has the startup time for the name server, with a 0 query
       count value.  */
 
    nrp = rinext+1;
    if( nrp == NUM_RATE_INFO_SLOTS ) nrp = 0;
    rate_info[nrp].itime = rate_info[rinext].itime;
    rate_info[nrp].qcount = 0;
 
    /* Save the current number of queries processed and the time. */
 
    rate_info[rinext].qcount = total_queries;
    rate_info[rinext].itime = current_time;
    sprintf( tstr, "idx = %4d itime = %d, qcount = %d", rinext,
         rate_info[rinext].itime, rate_info[rinext].qcount );
    write_log_entry( tstr );
 
    /* Update our pointers into the circular buffer. */
 
    rinext++;
    if( rinext == NUM_RATE_INFO_SLOTS ) {
        rinext = 0; riwrap = 1; }
    }

 
/*------------------------------------------------------------------------
--
--  calc_rate() - Find the query rate (in queries per second) between
--  two time bounds.
--
--  parameters:
--      start and end time of interval to be measured.
--
--  globals:
--      rate_info database and index value
--
--  return value:
--      float queries per second value
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
float calc_rate( start_time, end_time )
u_long start_time, end_time;
    {
    int rpp;
    u_long start_level, end_level;
    float rrate;

    /* Create a pointer to the last entry placed into the database. */

    rpp = rinext - 1;
    if( rpp < 0 ) rpp = NUM_RATE_INFO_SLOTS-1;

    /* If a time comes in set to 0 then substitute the oldest or most
       recent depending on which one. */

    if( start_time == 0 ) {
        if( riwrap == 0 ) start_time = rate_info[0].itime;
        else start_time = rate_info[rinext].itime;  }
    if( end_time == 0 ) {
        end_time = rate_info[rpp].itime; }
 
    /* Make sure a bogus call is not being made. */
 
    if( end_time <= start_time ) return( -1.0 );
 
    /* Look to see that the end time is not still in the future
       (as viewed from the rate info database. */
 
    if( end_time > rate_info[rpp].itime )
        end_time = rate_info[rpp].itime;

    /* Look to see that the start time does not precede the
       startup of the nameserver. */
 
    if( !riwrap && start_time < rate_info[0].itime )
        start_time = rate_info[0].itime;
    else if( riwrap && start_time < rate_info[rinext].itime )
        start_time = rate_info[rinext].itime;
 
    /* Get the query count level at each time point. */
 
    start_level = find_slot_and_interp( start_time );
    end_level = find_slot_and_interp( end_time );
 
    /* Return the number of queries per time interval. */
 
    rrate = (float)(end_level-start_level) /
            (float)(end_time-start_time);
    return( rrate );
    }

 
/*------------------------------------------------------------------------
--
--  find_slot_and_interp() - looks through rate database and finds the
--  interpolated query count at a given time.
--
--  parameters:
--      time of interest.
--
--  globals:
--      rate_info database and index value.
--
--  return value:
--      query count at the indicated time.
--      0 if there is a problem finding the time.
--
--  notes:
--      No error checking is done in this routine.  It is assumed that
--      the time fits within the time span in the database.  If this
--      is not the case the routine tries to return 0.
--
------------------------------------------------------------------------*/
u_long find_slot_and_interp( ptime )
u_long ptime;
    {
    int i, sp, prp, psp, soff;
    u_long level;
    float interp;

    /* Do some bounds checking. */

    if( !riwrap ) {
        if( ptime < rate_info[0].itime ) 
            return( rate_info[0].qcount );
        else if( ptime > rate_info[rinext-1].itime )
            return( rate_info[rinext-1].qcount ); 
        }
    else{
        if( ptime < rate_info[rinext].itime ) 
            return( rate_info[rinext].qcount );
        else { 
            prp = rinext - 1;
            if( prp < 0 ) prp = NUM_RATE_INFO_SLOTS-1;
            if( ptime > rate_info[prp].itime )
                return( rate_info[prp].qcount );
            }
        }
 
    /* For the case where we have wrapped the rate info database.
       Try to preselect a database location based on knowledge of the
       maintenance interval, and then loop backwards until the interval 
       time is less than the time, then loop forwards until we find the 
       transition. */
 
    if( riwrap == 1 ) {
        soff = 1 + (current_time - ptime) / maintenance_interval;
        sp = rinext - soff;
        if( sp < 0 ) sp = NUM_RATE_INFO_SLOTS + sp;
        for( i=0; i<NUM_RATE_INFO_SLOTS; i++ ) {
            if( rate_info[sp].itime < ptime ) break;
            sp--; if( sp < 0 ) sp = NUM_RATE_INFO_SLOTS-1; }
        if( i == NUM_RATE_INFO_SLOTS ) return( 0 );
        for( i=0; i<NUM_RATE_INFO_SLOTS; i++ ) {
            if( rate_info[sp].itime >= ptime ) break;
            sp++; if( sp == NUM_RATE_INFO_SLOTS ) sp = 0; }
        if( i == NUM_RATE_INFO_SLOTS ) return( 0 );
        }

    /* For the case where we are still starting to fill the database. */

    else {
        soff = 1 + (current_time - ptime) / maintenance_interval;
        sp = rinext - soff;
        if( sp < 0 ) sp = 0;
        if( sp >= rinext ) sp = rinext - 1;
        for( ; sp >= 0 ; sp-- ) if( rate_info[sp].itime <= ptime ) break;
        if( sp < 0 ) return( rate_info[0].qcount);
        for( ; sp < rinext; sp++ ) if( rate_info[sp].itime >= ptime ) break;
        if( sp == rinext ) return (rate_info[rinext-1].qcount);
        }

    /* If the time happened to be exactly on a boundary then just
       return that level. */

    if( rate_info[sp].itime == ptime ) return( (float)rate_info[sp].qcount );
 
    /* The sp variable now points to the second of two entries that
       straddle the desired start time.  Interpolate between their
       values to get the number of queries to be assigned to the
       period between start and end times. */
 
    psp = sp - 1;
    if( psp < 0 ) psp = NUM_RATE_INFO_SLOTS-1;
    interp = (float)(ptime - rate_info[psp].itime) /
        (float)(rate_info[sp].itime - rate_info[psp].itime);
    level = (float)rate_info[psp].qcount + 0.5 + (interp *
        (float)(rate_info[sp].qcount - rate_info[psp].qcount));
 
    /* Give the result back to the caller. */
 
    return( level );
    }

 
/*------------------------------------------------------------------------
--
--  create_rate_chart() - writes a simple usage chart of query counts
--  to the indicated file pointer.
--
--  parameters:
--      file pointer
--
--  globals:
--      rate_info database and index value.
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
#define VCNT 15  /* vertical size of rate chart (text lines)   */
#define SCNT 58  /* horozontal size of chart (columns)         */
#define BCNT 3   /* number of maintenance intervals per column */
 
void create_rate_chart( fp )
FILE *fp;
    {
    char chart[VCNT][SCNT+1];
    u_long cnts[SCNT];
    u_long min, max;
    float scale;
    int i, j, rp, val, rpp, trp;
    char lstr[16], rstr[16], tstr[100];
 
    /* Set up a pointer into the rate information, and clear out our
       chart data buffer. */
 
    rp = rinext - 1; if( rp<0 )  rp  = NUM_RATE_INFO_SLOTS-1;
    rpp = rp - 1;    if( rpp<0 ) rpp = NUM_RATE_INFO_SLOTS-1;
    memset( cnts, 0, sizeof(cnts) );
    trp = rp;
 
    /* Scan the preceding set of rate info records to create the data
       bins. */
 
    for( i=0; i<SCNT; i++ ) {
        for( j=0; j<BCNT; j++ ) {
            cnts[(SCNT-1)-i] +=
                (rate_info[rp].qcount-rate_info[rpp].qcount);
            rp = rpp;
            rpp--; if( rpp<0 ) rpp=NUM_RATE_INFO_SLOTS-1;
            if( rate_info[rpp].itime == 0 ) {j=BCNT;i=SCNT;}
            }
        }
 
    /* Scan the array for the minimum and maximum values. */
 
    min = 99999999;
    max = 0;
    for( i=0; i<SCNT; i++ ) {
        if( cnts[i] > max ) max = cnts[i];
        if( cnts[i] < min ) min = cnts[i];
        }
 
    /* Now create scaling factors that will let the count values be
       drawn into the chart. */
 
    scale = (float)VCNT / (float)(max-min);
    for( i=0; i<SCNT; i++ ) {
        cnts[i] = ((float)(cnts[i]-min) * scale) + 1.0;
        if( cnts[i] > VCNT ) cnts[i] = VCNT;
        }
 
    /* Now create the filled rate chart. */
 
    memset( chart, ' ', sizeof(chart) );
    for( i=0; i<VCNT; i++ ) chart[i][SCNT] = '\0';
    for( i=0; i<SCNT; i++ ) { for( j=0; j<cnts[i]; j++ ) chart[j][i]='X'; }

    /* Print out the header line with some statistics. */

    fprintf(fp,"Usage Rate Chart - Query counts per bin, %6.2f hours per bin\n",
        (float)(BCNT * maintenance_interval) / (float)3600 );
     
    /* Print out the chart with some labels. */
 
    for( i=VCNT-1; i>=0; i-- ) {
        val =  (float)((float)i/scale + (float)min + 0.5);
        fprintf( fp, "%5d | %s | %-5d\n", val, chart[i], val ); }
    memset( chart[0], '-', SCNT ); chart[0][SCNT] = '\0';
    fprintf( fp, "      +-%s-+\n", chart[0] );

    /* Create a string showing the left most time on the
       chart.  Make it blank if the chart doesn't go that
       far. */

    if( (BCNT*maintenance_interval*SCNT) >
        (rate_info[trp].itime-rate_info[rp].itime) ) {
        memset( lstr, ' ', sizeof(lstr) );
        lstr[12] = '\0'; }
    else { lstr[12] = '\0';
        strcpy( tstr, ctime( &rate_info[rp].itime ) );
        memcpy( lstr, &tstr[4], 12 ); }

    /* Make a string showing the most recent entry into the rate
       information database. */

    rstr[12] = '\0';
    strcpy( tstr, ctime( &rate_info[trp].itime ) );
    memcpy( rstr, &tstr[4], 12 );

    /* And actually print out the time scale */
 
    scale = (float)(rate_info[trp].itime - rate_info[rp].itime)/ 3600.0;
    fprintf( fp, "       ^%s", lstr );
    fprintf( fp, "       <--- %4.1f hours --->       %s^\n", scale, rstr );
    }
