/*-------------------------------------------------------------------------
--
--  pec_func.c - A set of functions to implement the proactive error
--  checking (PEC) features of the checker program.
--
--
===========================================================================
--                           UPDATE HISTORY                              --
--                                                                       --
-- VER     Date         Explanation of Changes              Author       --
---------------------------------------------------------------------------
--  0   |  1/8/93  | Initial release                 | Steve Miller - USC
--      |          |                                 |
-------------------------------------------------------------------------*/
 
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <netinet/in.h>
#include "checker.h"
 
extern QR_HTABLE *qr_tbl[QR_HASH_TABLE_SIZE];
 
extern u_long pec_test_length;
extern u_long pec_interval;
extern u_long pec_blocked_time;
extern u_long current_time;
extern u_long error_threshold;
extern u_long total_er_records;
extern u_long total_pc_records;
extern u_long total_ht_records;
extern FILE *ofp;

 
 
/*------------------------------------------------------------------------
--
--  scan_pec_state - a routine to check to see if it is time to select
--  a new PEC victim, or if it is time to stop a PEC test and report
--  on the results.
--
--  parameters:
--      none
--
--  globals:
--      current_time is used
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void scan_pec_state()
    {
    static PEC_CONTROL *pec_list = NULL;
    static u_long last_create_time = 0;
    PEC_CONTROL *pt;
 
    /* Scan the list of current tests to see if any are overdue for
       termination. */
 
    for( pt=pec_list; pt != NULL; pt=pt->next )
         if( pt->end_time < current_time ) pec_terminate( pt, &pec_list );

    /* If enough time has elapsed since the last test started then start
       a new one. */
 
    if( last_create_time+pec_interval < current_time ) {
        last_create_time = current_time;
        pec_start( &pec_list );
        }
 
    }

 
 
/*------------------------------------------------------------------------
--
--  pec_start - select a query name to be interdicted and create the
--  control structure to be associated with the test.
--
--  parameters:
--      list_head - pointer to the linked list of control structs
--
--  globals:
--      none
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void pec_start( list_head )
PEC_CONTROL **list_head;
    {
    QR_HTABLE *htpt;
    QR_DATA *qrpt;
    PEC_CONTROL *pcpt;
    FILE *lfp;
    char tstr[100];
 
    /* Select a query name to be used in the new test */
 
    htpt = pec_select();
    
    /* If no entry could be selected then issue a message to the log file
       that the loop safety limit was exceeded and this test opportunity
       has been skipped. */
       
    if( htpt == NULL ) {
        write_log_entry( "PEC testing opportunity missed." );
        return; }
 
    /* Mark that name's hash table entry as being under PEC testing. */
 
    htpt->pec_flag = PEC_TESTING;
 
    /* Log the fact that we have started a test on this name. */
 
    sprintf( tstr, "PEC started on %s/%d", htpt->q_name, htpt->q_type );
    write_log_entry( tstr );
 
    /* Create the control structure and set the test's start
       and end times. */
 
    pcpt = (PEC_CONTROL *)calloc( 1, sizeof(PEC_CONTROL) );
    pcpt->start_time = current_time;
    pcpt->end_time = current_time + pec_test_length;
    pcpt->pec_victim = htpt;
    total_pc_records++;
 
    /* Loop through all of the IP addresses that have made reference to
       this query name making copies of their error statistics.  The copies
       will be used to generate a report of activity during the PEC testing
       period. */
 
    pec_copy_stats( pcpt );
 
    /* Insert the control structure in the link list. */
 
    pcpt->next = *list_head;
    *list_head = pcpt;
    }

 
 
/*------------------------------------------------------------------------
--
--  pec_terminate - stop and report on the test controlled by the
--  structure pointed to by the incoming parameter.
--
--  parameters:
--      pt - pointer to the control structure for the test to be ended
--      list_head - pointer to the linked list of control structures
--
--  globals:
--      none
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void pec_terminate( pt, list_head )
PEC_CONTROL *pt;
PEC_CONTROL **list_head;
    {
    PEC_CONTROL *tpt, **otp;
    char tstr[100];
 
    /* If the end time was for the waiting period after a test run then
       clean up the control structure and release the name for a possible
       retest if the selection mechanism should choose to do so. */
 
    if( pt->pec_victim->pec_flag == PEC_RECENTLY_TESTED ) {
 
        /* Set the pec_flag to PEC_NOT_TESTING */
 
        pt->pec_victim->pec_flag = PEC_NOT_TESTING;
 
        /* Find the link that points to this control structure */
 
        for( tpt=*list_head,otp=list_head; tpt != NULL; tpt=tpt->next ) {
            if( tpt == pt ) break;
            else otp = &tpt->next;
            }
 
        /* Unlink this control structure from the list. */
 
        *otp = pt->next;
 
        /* Free up the control structure and decrement the record count */
 
        free( pt );
        total_pc_records--;
        }
 
 
    /* If the end time was for an active test then issue reports, stop
       the interception of responses for this query name, and block the
       name from further tests for awhile. */
 
    else if( pt->pec_victim->pec_flag == PEC_TESTING ) {
 
        /* Issue PEC report records on all IP addresses using the query
           name pointed to in the control structure. */
 
        pec_report( pt );
 
        /* Set the pec test flag in the QR_HTABLE structure to show that
           this name has been recently tested. */
 
        pt->pec_victim->pec_flag = PEC_RECENTLY_TESTED;
 
        /* Set the end time to be current time+pec_blocked_time to show that
           this name should not be retested for awhile. */
 
        pt->end_time = current_time + pec_blocked_time;
 
        /* Log that the pec test is over */
 
        sprintf( tstr, "PEC stopped on %s/%d", pt->pec_victim->q_name,
            pt->pec_victim->q_type );
        write_log_entry( tstr );
 
        }
    }

 
 
/*------------------------------------------------------------------------
--
--  pec_select - choose a query name to be tested using the PEC system.
--
--  parameters:
--      none
--
--  globals:
--      none
--
--  return value:
--      pointer to the QR_HTABLE structure that heads the chosen
--          branch of the database
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
QR_HTABLE *pec_select()
    {
    u_long i,j,k,m;
    QR_HTABLE *op;
    char tstr[100];
 
    /* Loop through the selection process until a valid record is
       found, or we exceed a safety limit. */
 
    for( i=0, op=NULL; op == NULL && i < PEC_SAFETY_LIMIT; i++ ) {
 
        /* Create a random number between 1 and the number of QR_HTABLE
           records in the database at this moment. */
 
        k = ((u_long)random() % total_ht_records) + 1;
 
        /* Iterate down the hash list until the selected QR_HTABLE record
           is found. */
 
        for( j=0; j<QR_HASH_TABLE_SIZE && k>0; j++ ) {
            for( op=qr_tbl[j]; op!=NULL; op=op->hclink ) { 
                 k--; if(!k) break;
                 }
            }
        
        /* If the selected record is not under test and not blocked
           then use it. */
 
        if( op == NULL ) continue;
        else if( op->pec_flag == PEC_NOT_TESTING ) break;
        else { op = NULL; }
        }
 
    /* Return the selected record or NULL to indicate that we hit the
       loop safety limit. */
 
    return( op );
    }

 
 
/*------------------------------------------------------------------------
--
--  pec_report - dump one of two kinds of report, and then clear out all
--  error records associated with all IP addresses that have referenced
--  the query name pointed to by the input parameter.
--
--  parameters:
--      pt - pointer to PEC_CONTROL structure
--
--  globals:
--      none
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void pec_report( pt )
PEC_CONTROL *pt;
    {
    QR_HTABLE *htpt;
    QR_DATA *qrpt;
    ERR_DATA saverr, olderr;
    u_long i, tet, tmp;
    char tstr[100];
 
    /* Get a simple pointer to the query name head */
 
    htpt = pt->pec_victim;
 
    /* Set the error threshold based on the type of report being
       generated, and generate a report header that is specific
       to PEC processing. */
 
    tet = 1;
    dump_pec_header( ofp, htpt->q_name, htpt->q_type );
 
    /* Do reports on all IP addresses that have made references
       to this query name, and then make sure those addresses have
       saved error records. */
 
    for( qrpt=htpt->qrlink; qrpt != NULL; qrpt=qrpt->qrk.flink ) {
 
        /* Save the contents of the IP address' error record and put just
           the deltas there instead.  We will replace the contents after
           the report is finished. */
 
        memset( &saverr, 0, sizeof(saverr) );
        if( qrpt->qrk.iprec->err_data != NULL )
            memcpy( &saverr, qrpt->qrk.iprec->err_data, sizeof(saverr) );
        memset( &olderr, 0, sizeof(olderr) );
        if( qrpt->qrk.iprec->pec_save != NULL )
            memcpy( &olderr, qrpt->qrk.iprec->pec_save, sizeof(olderr) );
        for( i=0; i<NUMBER_OF_ERROR_TYPES; i++ ) {
            qrpt->qrk.iprec->err_data->ea[i].count =
                 saverr.ea[i].count - olderr.ea[i].count;
            qrpt->qrk.iprec->err_data->ea[i].ftime = pt->start_time;
            qrpt->qrk.iprec->err_data->ea[i].ltime = pt->end_time; 
            }
 
        /* Calculate the delta query statistics and save the current counts. */
 
        tmp = qrpt->qrk.iprec->tq - qrpt->qrk.iprec->sav_tq;
        qrpt->qrk.iprec->sav_tq = qrpt->qrk.iprec->tq;
        qrpt->qrk.iprec->tq = tmp;
        tmp = qrpt->qrk.iprec->tr - qrpt->qrk.iprec->sav_tr;
        qrpt->qrk.iprec->sav_tr = qrpt->qrk.iprec->tr;
        qrpt->qrk.iprec->tr = tmp;
        tmp = qrpt->qrk.iprec->neq - qrpt->qrk.iprec->sav_neq;
        qrpt->qrk.iprec->sav_neq = qrpt->qrk.iprec->neq;
        qrpt->qrk.iprec->neq = tmp;
        tmp = qrpt->qrk.iprec->ner - qrpt->qrk.iprec->sav_ner;
        qrpt->qrk.iprec->sav_ner = qrpt->qrk.iprec->ner;
        qrpt->qrk.iprec->ner = tmp;
 
        /* Output the error record for this IP address */
 
        dump_err_record( ofp, qrpt->qrk.iprec, tet );
 
        /* Bring the saved error record back and delete the one created
           for the PEC testing interval. */
 
        if( qrpt->qrk.iprec->err_data != NULL )
            memcpy( qrpt->qrk.iprec->err_data, &saverr, sizeof(saverr) );
        if( qrpt->qrk.iprec->pec_save != NULL ) {
            total_er_records--;
            free( qrpt->qrk.iprec->pec_save );
            qrpt->qrk.iprec->pec_save = NULL; }
 
        /* Copy the query/response statistics back from the save area. */
 
        qrpt->qrk.iprec->tq = qrpt->qrk.iprec->sav_tq;
        qrpt->qrk.iprec->tr = qrpt->qrk.iprec->sav_tr;
        qrpt->qrk.iprec->neq = qrpt->qrk.iprec->sav_neq;
        qrpt->qrk.iprec->ner = qrpt->qrk.iprec->sav_ner;
        }
    }

 
 
/*------------------------------------------------------------------------
--
--  pec_copy_stats - Loop through all of the IP addresses that have made
--  reference to the query name/type to be PEC tested and copies their
--  error statistics over to a save area so that deltas occuring during
--  the PEC testing period can be reported.
--
--  parameters:
--      pt - pointer to a PEC_CONTROL structure
--
--  globals:
--      none
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void pec_copy_stats( pt )
PEC_CONTROL *pt;
    {
    QR_HTABLE *htpt;
    QR_DATA *qrpt;
 
    /* Get a simple pointer to the query name head */
 
    htpt = pt->pec_victim;
 
    /* Loop through all of the IP addresses that have made reference
       to the query name. */
 
    for( qrpt=htpt->qrlink; qrpt != NULL; qrpt=qrpt->qrk.flink ) {
 
        /* If there is an error record then allocate a new one and
           copy the current error record contents over there. */
 
        if( qrpt->qrk.iprec->err_data != NULL ) {
            qrpt->qrk.iprec->pec_save = 
                (ERR_DATA *)calloc( 1, sizeof(ERR_DATA) );
            memcpy( qrpt->qrk.iprec->pec_save, qrpt->qrk.iprec->err_data,
                sizeof( ERR_DATA ) );
            total_er_records++;
            }
 
        /* Copy the query/response statistics over to the save area. */
 
        qrpt->qrk.iprec->sav_tq = qrpt->qrk.iprec->tq;
        qrpt->qrk.iprec->sav_tr = qrpt->qrk.iprec->tr;
        qrpt->qrk.iprec->sav_neq = qrpt->qrk.iprec->neq;
        qrpt->qrk.iprec->sav_ner = qrpt->qrk.iprec->ner;
        }
    }

