/*-------------------------------------------------------------------------
--
--  reports.c - A set of routines to generate reports on checker's
--  database contents.
--
--
===========================================================================
--                           UPDATE HISTORY                              --
--                                                                       --
-- VER     Date         Explanation of Changes              Author       --
---------------------------------------------------------------------------
--  0   | 11/30/92 | Initial version                 | Steve Miller - USC   
--      |          |                                 |
-------------------------------------------------------------------------*/
 
#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <arpa/nameser.h>
#include "checker.h"
 
extern int failflag;
extern int memfault;
extern FILE *ofp;
 
extern QR_DATA  timeq[TIME_QUEUE_SLOT_COUNT];
extern IP_DATA   *ip_tbl[IP_HASH_TABLE_SIZE];
extern QR_HTABLE *qr_tbl[QR_HASH_TABLE_SIZE];
 
extern u_long total_ht_records;  /* QR_HTABLE */
extern u_long total_qr_records;  /* QR_DATA   */
extern u_long total_ip_records;  /* IP_DATA   */
extern u_long total_dm_records;  /* IPD_DATA  */
extern u_long total_er_records;  /* ERR_DATA  */
extern u_long total_qn_chars;    /* query name string storage */
extern u_long total_packets;     /* total packets for chosen server */
extern u_long total_queries;     /* total queries actually processed */
extern u_long total_responses;   /* total responses actually processed */
 
extern char   output_file_name[100];
extern char   logging_file_name[100];
extern char   dig_path[100];
extern u_long qr_deletion_time;
extern u_long nocheck_authority;
extern u_long echeck_mask;
extern u_long ck_debug_flag;
extern u_long translate_address;
extern u_long disable_tq_flag;
extern u_long error_threshold;
extern u_long pec_checking;
extern u_long pec_blocked_time;
extern u_long pec_test_length;
extern u_long pec_interval;
extern u_long memory_logging;
extern u_long local_domain_checking;
extern u_long port_53_checking;
extern u_long maintenance_interval;

extern u_long current_time;
extern u_long active_report;
extern u_long report_pid;
extern char erstr[NUMBER_OF_ERROR_TYPES][ERR_STR_LENGTH+1];
 
typedef struct { int in; int out } PIPEFD;
static PIPEFD pipefd;
int err_header_flag = 0;
char err_header_string[500];
 
#define NUM_QTNAM 21
char *qtype_names[NUM_QTNAM] = {
"unkwn", "A    ", "NS   ", "MD   ", "MF   ", "CNAME",
"SOA  ", "MB   ", "MG   ", "MR   ", "NULL ", "WKS  ",
"PTR  ", "HINFO", "MINFO", "MX   ", "TXT  ",
"AXFR ", "MAILB", "MAILA", "*    " };
 
#define NUM_RCODE 7
char *rcode[NUM_RCODE] = { "OK      ",
"formerr ", "servfail", "nxdomain", "not impl", "refused ", "bad code" };

 
 
/*------------------------------------------------------------------------
--
--  dump_tables - a little routine to output a textual list of the
--  contents of checker()'s database.
--
--  parameters:
--      none
--
--  globals:
--      The query/response, ip address, and time queue tables are read.
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void dump_tables( )
    {
    struct timeval tm;
    static struct timeval report_time= { 0, 0 };
    char tstr[100];
    extern int errno;
    void sigbus_hdlr();
    void sigsegv_hdlr();
 
    /* If there is a report being processed then ignore the request
       for a new one. */
 
    if( active_report ) {
        if( report_time.tv_sec+REPORT_TIMEOUT < current_time ) {
            kill( report_pid, 9 );
            write_log_entry( "report timed out, kill attempted" );
            close( pipefd.in );
            close( pipefd.out );
            active_report = 0; }
        else {
            write_log_entry( "report in progress, request ignored." );
            return; }
        }
 
    /* Set the active report flag, and retain the current time so that
       we can check that it hasn't gotten stuck. */
 
    active_report = 1;
    gettimeofday( &report_time, (struct timezone *)NULL );
    pipe( &pipefd );
 
    /* Now split the processing thread and return if we are the main
       parent process. */
 
    if( (report_pid=fork()) > 0 ) {
        sprintf( tstr, "report child process id is %d", report_pid );
        write_log_entry( tstr );
        return; }
    else if( report_pid < 0 ) {
        sprintf( tstr, "report child create failed, err=%d", errno );
        write_log_entry( tstr );
        active_report = 0; 
        return; }
 
    /* We are now the child process, install some signal handlers in
       case something dies and lower our priority for a while. */
 
    signal( SIGBUS, sigbus_hdlr );
    signal( SIGSEGV, sigsegv_hdlr );
    nice( 10 );
    fprintf( ofp, "\n" );
 
    /* Do the various phases of the report. */
 
    dump_stats( ofp );   
    dump_ipdata( ofp, 0 );
    dump_ipdata( ofp, 1 );
    dump_qrdata( ofp );
    dump_errors( ofp );
    finish_report();
    }

 
 
/*------------------------------------------------------------------------
--
--  dump_stats - a little routine to output the record and transaction
--  counts for the current database.
--
--  parameters:
--      file pointer
--
--  globals:
--      The query/response and ip address tables are read.
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void dump_stats( fp )
FILE *fp;
    {
    u_long total_qrht_entries, total_ipht_entries;
    u_long total_ip_hcoll, total_qr_hcoll;
    u_long i, calcd_db_size;
    time_t ttm;
 
    /* Scan the two hash queues to get the number of slots that
       actually have entries. */
 
    total_ipht_entries = total_ip_hcoll = 0;
    total_qrht_entries = total_qr_hcoll = 0;
 
    for( i=0; i<IP_HASH_TABLE_SIZE; i++ ) {
        if( ip_tbl[i] != NULL ) { 
            total_ipht_entries++; 
            if( ip_tbl[i]->flink != NULL ) total_ip_hcoll++; }
        }
    for( i=0; i<QR_HASH_TABLE_SIZE; i++ ) {
        if( qr_tbl[i] != NULL ) {
            total_qrht_entries++; 
            if( qr_tbl[i]->hclink != NULL ) total_qr_hcoll++; }
        }
 
    /* Calculate the size of the database at this point. */
 
    calcd_db_size = calc_db_size();
 
    /* Get the current time */
 
    ttm = time( (time_t *)NULL );
 
    /* Print out a title and the packets processed statistics */
 
    fprintf( fp, "\f\n" );
    fprintf( fp, "CHECKER PROCESSING STATISTICS at %s\n", ctime(&ttm) );
    fprintf( fp, "%d packets seen; %d discarded (%6.2f percent)\n",
             total_packets, total_packets-(total_responses+total_queries),
             (float)(100.0*(total_packets-(total_responses+total_queries))/
             total_packets) );
    fprintf( fp, "%d queries (%6.2f percent), %-6.2f per hour\n",
             total_queries, (float)(100.0*total_queries/total_packets),
             3600.0*calc_rate( 0, 0 ) );
    fprintf( fp, "%d responses (%6.2f percent)\n", total_responses,
             (float)(100.0*total_responses/total_packets) );
    fprintf( fp, "Query table (%d slots), ", QR_HASH_TABLE_SIZE );
    fprintf( fp, "%-6.2f pct. filled, %6.3f pct. collisions\n",
             (float)(100.0*total_qrht_entries/QR_HASH_TABLE_SIZE),
             (float)(100.0*total_qr_hcoll/QR_HASH_TABLE_SIZE) );
    fprintf( fp, "IP address table (%d slots), ", IP_HASH_TABLE_SIZE );
    fprintf( fp, "%-6.2f pct. filled, %6.3f pct. collisions\n\n",
             (float)(100.0*total_ipht_entries/IP_HASH_TABLE_SIZE),
             (float)(100.0*total_ip_hcoll/IP_HASH_TABLE_SIZE) );

    fprintf( fp, "nocheck_authority: %d               ck_debug_flag: %d\n",
             nocheck_authority, ck_debug_flag );
    fprintf( fp, "translate_address: %d             disable_tq_flag: %d\n",
             translate_address, disable_tq_flag );
    fprintf( fp, "pec_checking:      %d       local_domain_checking: %d\n",
             pec_checking, local_domain_checking );
    fprintf( fp, "memory_logging:    %d            port_53_checking: %d\n\n",
             memory_logging, port_53_checking );                                    

    fprintf( fp, "%d bytes in dynamic database (by calculation)\n",
         calcd_db_size );
    fprintf( fp, "%-6.2f percent of ip addresses have error records\n\n",
             (float)(100.0*total_er_records/total_ip_records) );

    fprintf( fp,
         "%-5d QR_HTABLE records,%7d bytes, %6.2f percent of database\n",
         total_ht_records,
         total_ht_records*sizeof(QR_HTABLE),
         100.0 * (float)(total_ht_records*sizeof(QR_HTABLE)) / 
         (float)calcd_db_size );
    fprintf( fp,
         "%-5d QR_DATA records,  %7d bytes, %6.2f percent of database\n",
         total_qr_records, total_qr_records*sizeof(QR_DATA),
         (float)(100.0*(total_qr_records*sizeof(QR_DATA))/calcd_db_size) );
    fprintf( fp,
         "%-5d IP_DATA records,  %7d bytes, %6.2f percent of database\n",
         total_ip_records, total_ip_records*sizeof(IP_DATA),
         (float)(100.0*(total_ip_records*sizeof(IP_DATA))/calcd_db_size) );
    fprintf( fp,
         "%-5d ERR_DATA records, %7d bytes, %6.2f percent of database\n",
         total_er_records, total_er_records*sizeof(ERR_DATA),
         (float)(100.0*(total_er_records*sizeof(ERR_DATA))/calcd_db_size) );
    fprintf( fp,
         "%-5d IPD_DATA records, %7d bytes, %6.2f percent of database\n",
         total_dm_records, total_dm_records*sizeof(IPD_DATA),
         (float)(100.0*(total_dm_records*sizeof(IPD_DATA))/calcd_db_size) );
    fprintf( fp,
         "      name strings,     %7d bytes, %6.2f percent of database\n\n",
         total_qn_chars,
         100.0 * ((float)total_qn_chars / (float)calcd_db_size) );

    fprintf( fp, "%d secs for negative cache, %d slots in delete queue\n",
             NEGATIVE_TTL, TIME_QUEUE_SLOT_COUNT );
    fprintf( fp, "%d secs expected for requery, %d secs for processing\n",
             STDTIMOUT, DELTA );
    fprintf( fp, 
             "recursion lb window is %d secs, servfail window is %d secs\n",
             LBUCKET_WINDOW, IPQ_LBKT_WINDOW );
    fprintf( fp, "%d entries trigger a recursion error (%d slots max)\n",
             MAX_LB_COUNT, MAX_STACK );
    fprintf( fp, "%d entries trigger a bad server response error (%d max)\n\n",
             MAX_IPQ_COUNT, MAX_IPQ_STACK );
 
    fprintf( fp, "output report filename is %s\n", output_file_name );
    fprintf( fp, "activity log file is %s\n", logging_file_name );
    fprintf( fp, "path to DIG name translation program is %s\n", dig_path );
    fprintf( fp, "%d secs per maintenance interval, %-6.2f hours per cycle\n",
             maintenance_interval,
             (float)(maintenance_interval*TIME_QUEUE_SLOT_COUNT/3600.0));
    fprintf( fp, "unreferenced query records last for %d seconds\n",
             qr_deletion_time );
    fprintf( fp, "error check mask is 0x%08X, with a %d threshold\n",
             echeck_mask, error_threshold );
    fprintf( fp, "%d secs between PEC starts, %d secs is PEC test time\n",
             pec_interval, pec_test_length );
    fprintf( fp, "names are blocked after pec checks for %d seconds\n\n",
             pec_blocked_time );

    create_rate_chart( fp );
    fflush( fp );
    }

 
 
/*------------------------------------------------------------------------
--
--  dump_errors - a little routine to output a textual list of the
--  error records in  checker()'s database.
--
--  parameters:
--      file pointer
--
--  globals:
--      The query/response, ip address, and time queue tables are read.
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void dump_errors( fp )
FILE *fp;
    {
    u_long i, j, k;
    QR_HTABLE *qrhp;
    ERR_DATA *ep;
    QR_DATA *qrp;
    IP_DATA *ipp;
    char temp[200];
    char tstr[30], fstr[15], lstr[15];
    IP_DATA *slots[ERR_DUMP_SLOTS];
    float ccnt, counts[ERR_DUMP_SLOTS];

    /* Dump out a heading for the error occurance listing. */
    
    fprintf( fp, "\f\n" );
    dump_err_header( fp );
    memset( slots, 0, sizeof(slots) );
    memset( counts, 0, sizeof(counts) );
 
    /* Loop through the hash table contents. */
 
    for( i=0; i<IP_HASH_TABLE_SIZE; i++ ) {
 
        /* Loop down the linked list of entries at this hash table slot */
 
        for( ipp=ip_tbl[i]; ipp != NULL; ipp=ipp->flink ) {

            /* Generate the sort list entry. */

            if( ipp->err_data != NULL ) {

                /* Look through the error entries to find the one with the
                   greatest count. */

                for( ccnt=0.0,j=0; j < NUMBER_OF_ERROR_TYPES; j++ ) {
                    if( (float)ipp->err_data->ea[j].count > ccnt ) {
                        ccnt = ipp->err_data->ea[j].count; k=j; }
                    }

                /* If this error count does not meet our minimum threshold
                   then go to the next record. */

                if( ccnt < (float)error_threshold ) continue;

                /* Calculate the error occurance rate. */

                if( ipp->err_data->ea[k].ftime ==
                    ipp->err_data->ea[k].ltime ) ccnt = 0.0;
                else ccnt = 3600.0*ccnt / (float)(ipp->err_data->ea[k].ltime-
                    ipp->err_data->ea[k].ftime);

                /* Place that rate into the sort list. */

                if( ccnt > counts[ERR_DUMP_SLOTS-1] ) {
                    for( j=0; j<ERR_DUMP_SLOTS-1; j++ ) {
                        if( counts[j] < ccnt ) {
                            bcopy( &counts[j], &counts[j+1],
                                    ((ERR_DUMP_SLOTS-j)-1)*sizeof(float) );
                            bcopy( &slots[j], &slots[j+1],
                                    ((ERR_DUMP_SLOTS-j)-1)*sizeof(IP_DATA*) );
                            counts[j] = ccnt;
                            slots[j] = ipp;
                            break;
                            }
                        }
                    }
                }
            }
        }

    /* Now that we have an ordered list of the worst offenders, dump
       each of the error records into the report file. */

    for( i=0; i<ERR_DUMP_SLOTS && slots[i] != NULL; i++ )
        dump_err_record( fp, slots[i], error_threshold );
    }

 
 
/*------------------------------------------------------------------------
--
--  dump_err_record() - a routine to dump out the contents of an error
--  record.
--
--  parameters:
--      error record pointer
--
--  globals:
--      none
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void dump_err_record( fp, ipp, thold )                
FILE *fp;
IP_DATA *ipp;
u_long thold;
    {
    u_long j;    
    ERR_DATA *ep;
    char temp[256], stmp[50];
    char tstr[30], fstr[15], lstr[15];

    /* Make up a pointer to make the code a little more readable, and
       check to see if there is an error record.  No error record means
       that this node has been behaving exceptionally well. */
 
    ep = ipp->err_data;
    if( ep == NULL ) return;
 
    /* Loop through each slot to make sure that there is one
                   entry above the theshold set by the user. */
 
    for( j=0; j<NUMBER_OF_ERROR_TYPES; j++ )
        if( ep->ea[j].count >= thold ) break;
        if( j == NUMBER_OF_ERROR_TYPES ) return;
 
    /* If the header string flag is set then this must be the first record
       for this report, dump the header also. */
 
    if( err_header_flag ) {
        fprintf( fp, "%s", err_header_string );
        err_header_flag = 0; }

    /* Create a little text string showing whether or not this machine
       supports a server or just simple resolvers. */

    if(      (ipp->srcflags&0x3) == 1 ) strcpy( stmp, "(server)" );
    else if( (ipp->srcflags&0x3) == 2 ) strcpy( stmp, "(resolver)" );
    else if( (ipp->srcflags&0x3) == 3 ) strcpy( stmp, "(server and resolver)" );
 
    /* Dump out the remote node's address, and name if requested. */
 
    if( translate_address ) {
        if( ipp->resolved_name == NULL ){
            xlate_addr( ipp->remote_ip_addr, temp, sizeof(temp) );
            temp[sizeof(temp)-1] = '\0';
            ipp->resolved_name = (char *)calloc( 1, strlen(temp)+1 );
            strcpy( ipp->resolved_name, temp );
            }
        fprintf( fp, "\n%s (%s) %s\n", decode_addr(ipp->remote_ip_addr),
            ipp->resolved_name, stmp ); }
    else {
        fprintf( fp, "\n%s %s\n", decode_addr(ipp->remote_ip_addr), stmp ); }
 
    /* Dump out the query statistics for this particular node. */
 
    fprintf( fp, "%d queries,   %6.2f/hour, %6.2f percent with errors\n",
        ipp->tq,
        3600.0*(float)ipp->tq/(float)(ipp->q_latest_time-ipp->q_first_time),
        100.0 * (float)(ipp->tq-ipp->neq) / (float)ipp->tq );
    fprintf( fp, "%d responses, %6.2f/hour, %6.2f percent with errors\n",
        ipp->tr,
        3600.0*(float)ipp->tr/(float)(ipp->q_latest_time-ipp->q_first_time),
        100.0 * (float)(ipp->tr-ipp->ner) / (float)ipp->tr );
 
    /* Loop through each error slot and print out the count and
       name, and times for each that is nonzero. */
 
    for( j=0; j<NUMBER_OF_ERROR_TYPES; j++ ) {
        if( ep->ea[j].count != 0 ) {
            fprintf( fp, "      %s   %-5ld", erstr[j], ep->ea[j].count );
            if( ep->ea[j].ftime == ep->ea[j].ltime ) 
               fprintf( fp, " (           ," );
            else
               fprintf( fp, " (%-6.2f/hour,", 3600.0 * (float)ep->ea[j].count /
                   (float)(ep->ea[j].ltime - ep->ea[j].ftime) );
            fprintf( fp, "%-6.2f hours,",
               (float)(ep->ea[j].ltime - ep->ea[j].ftime) / 3600.0 );
            fstr[12] = '\0';
            strcpy( tstr, ctime( &ep->ea[j].ftime ) );
            memcpy( fstr, &tstr[4], 12 );
            fprintf( fp, "%s)\n", fstr );
            }
        }
    fflush( fp );
    }

 
 
/*------------------------------------------------------------------------
--
--  dump_err_header() - a routine to dump out the error report header
--
--  parameters:
--      file pointer
--
--  globals:
--      none
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void dump_err_header( fp )
FILE *fp;
    {
    time_t ttm;
    char temp[500];
 
    /* Get the current time */
 
    ttm = time( (time_t *)NULL );
 
    /* Dump out a heading for the error occurance listing. */
 
    sprintf( err_header_string, 
    "\n--------------------------------------------------------------------\n");
    sprintf( temp,
    "ERROR TABLE, threshold=%d, %s", error_threshold, ctime(&ttm) );
    strcat( err_header_string, temp );
    sprintf( temp,
    "--------------------------------------------------------------------\n" );
    strcat( err_header_string, temp );
    err_header_flag = 1;
    }

 
 
/*------------------------------------------------------------------------
--
--  dump_pec_header() - a routine to dump out the error report header
--  after a PEC test run
--
--  parameters:
--      file pointer
--
--  globals:
--      none
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void dump_pec_header( fp, name, type )
FILE *fp;
char *name;
u_long type;
    {
    time_t ttm;
    char temp[500];
    int k;
 
    /* Get the current time */
 
    ttm = time( (time_t *)NULL );
 
    /* Dump out a heading for the error occurance listing. */
 
    k = type;
    if( k >= NUM_QTNAM ) k = NUM_QTNAM-1;
    if( k < 0 ) k = 0;
 
    sprintf( err_header_string,
    "\n--------------------------------------------------------------------\n");
    sprintf( temp,
    "PEC REPORT for %s, type %s, %s", name, qtype_names[k], ctime(&ttm) );
    strcat( err_header_string, temp );
    sprintf( temp,
    "--------------------------------------------------------------------\n" );
    strcat( err_header_string, temp );
    err_header_flag = 1;
    }

 
 
/*------------------------------------------------------------------------
--
--  dump_ipdata - a little routine to output a textual list of the
--  IP addresses with the greatest number of different queries.
--
--  parameters:
--      file pointer
--      sort data type: 0 = reference count, 1 = total queries
--
--  globals:
--      The query/response, ip address, and time queue tables are read.
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void dump_ipdata( fp, dtype )
FILE *fp;
int dtype;
    {
    u_long i, j, f, t1, t2, tx, tcnt;
    IP_DATA *ipp;
    float ip_total;
    IP_DATA *slots[IP_DUMP_SLOTS];
    float ccnt, counts[IP_DUMP_SLOTS];
    u_long rates[IP_DUMP_SLOTS];
    char tqn[300], temp[256];
    time_t ttm;
 
    /* Get the current time */
 
    ttm = time( (time_t *)NULL );
 
    /* Clear out the sort array */
 
    memset( slots, 0, sizeof(slots) );
    memset( rates, 0, sizeof(rates) );
    memset( counts,0, sizeof(counts));
 
    /* Loop through the hash table contents. */
 
    for( i=0; i<IP_HASH_TABLE_SIZE; i++ ) {
 
        /* Loop through all entries at this hash table slot */
 
        for( ipp=ip_tbl[i]; ipp != NULL; ipp=ipp->flink ) {
 
            /* If the report is on the number of different queries
               per IP address then get the reference count as the item
               to be sorted on. */
 
            if( dtype == 0 ) {
                ccnt = 100.0 * (float)ipp->reference_count /
                    (float)total_qr_records;
                tcnt = ipp->reference_count; }
 
            /* If the report is on the rate of query activity then
               get the client's query rate during its active period
               and divide that by Checker's total rate during that
               time period.  If the period of interest is less than
               three maintenance intervals then stretch it to at least
               that width. */
 
            else if( dtype == 1 ) {
               if( ipp->q_latest_time == ipp->q_first_time ) ccnt = 0.0;
               else {
                   t1 = ipp->q_first_time; t2 = ipp->q_latest_time;
                   if( (t2-t1) < 3*maintenance_interval )
                       tx = ((3*maintenance_interval - (t2-t1)) / 2) + 1;
                   else tx = 0;
                   t1 -= tx; t2 += tx;
                   ccnt = 100.0 * ((float)ipp->tq / 
                          (float)(t2 - t1)) / calc_rate( t1, t2 ); 
                   }
               tcnt = ipp->tq;
               }
 
            /* Put the new entry into the sort table if it is greater
               than the least entry currently there.  To actually
               place the entry the list is searched for the correct
               point and then every thing is moved to make room, with
               the least entry being popped off of the list. */
 
            if( ccnt > counts[IP_DUMP_SLOTS-1] ) {
                for( j=0; j<IP_DUMP_SLOTS-1; j++ ) {
                    if( counts[j] < ccnt ) {
                        bcopy( &counts[j], &counts[j+1],
                                ((IP_DUMP_SLOTS-j)-1)*sizeof(float) );
                        bcopy( &slots[j], &slots[j+1],
                                ((IP_DUMP_SLOTS-j)-1)*sizeof(IP_DATA*) );
                        bcopy( &rates[j], &rates[j+1],
                                ((IP_DUMP_SLOTS-j)-1)*sizeof(u_long) );
                        counts[j] = ccnt;
                        slots[j] = ipp;
                        rates[j] = tcnt;
                        break;
                        }
                    }
                }
            }
        }
 
    /* Dump out a title and the table header for the unique query table. */
 
    if( dtype == 0 ) {
        for( i=0,ip_total=0.0; i<IP_DUMP_SLOTS; i++ ) ip_total += counts[i];
        if( total_ip_records < IP_DUMP_SLOTS ) i = total_ip_records;
        fprintf( fp, "\f\n" );
        fprintf( fp,"UNIQUE QUERIES by SOURCE ADDRESS at %s\n",ctime(&ttm));
        fprintf( fp,
            "(ipaddr's on list supplied %6.2f percent of total queries)\n",
            ip_total );
        fprintf( fp,
            "(ipaddr's on list are %6.2f percent of addresses stored)\n",
            (float)(100.0*i/total_ip_records) );
        fprintf( fp,
            "(pct is client's percent of total query names stored)\n" );
        fprintf( fp,
            "(count is client's total of unique queries in db)\n\n" );
        fprintf( fp,
        " pct    count     ipaddr             ipaddr DNS name\n" );
        fprintf( fp,
        "----------------------------------------------------------------\n");
        }
 
    /* Dump out a title and the table header for the query rate table. */
 
    else if( dtype == 1 ) {
        i = IP_DUMP_SLOTS;
        if( total_ip_records < IP_DUMP_SLOTS ) i = total_ip_records;
        fprintf( fp, "\f\n" );
        fprintf( fp,"QUERY RATE by SOURCE ADDRESS at %s\n",ctime(&ttm));
        fprintf( fp,"(list is %6.2f percent of ip addresses stored)\n",
                 (float)( 100.0 * i / total_ip_records ) );
        fprintf( fp,
        "(pct is ipaddr's percent of total over its active period)\n" );
        fprintf( fp,
        "(count is ipaddr's total number of queries)\n\n" );
        fprintf( fp,
        " pct    count     ipaddr                    name\n" );
        fprintf( fp,
        "----------------------------------------------------------------\n");
        }
 
    /* Loop through the sort table contents. */
 
    for( i=0; i<IP_DUMP_SLOTS && slots[i] != NULL; i++ ) {
 
        /* If the user has requested us to then translate the IP addresses
           into DNS names. */
 
        if( translate_address ) {
 
            /* only translate the name if it hasn't already been done. */
 
            if( slots[i]->resolved_name == NULL ){
                xlate_addr( slots[i]->remote_ip_addr, temp, sizeof(temp) );
                temp[sizeof(temp)-1] = '\0';
                slots[i]->resolved_name = (char *)calloc( 1, strlen(temp)+1 );
                strcpy( slots[i]->resolved_name, temp );
                }

            /* Put the IP address in first, then the DNS name. */
 
            strcpy( tqn, decode_addr(slots[i]->remote_ip_addr) );
            strcat( tqn, "               " );
            tqn[17] = '\0'; 
            strcat( tqn, slots[i]->resolved_name );
            }
 
        /* If no translation was requested then just put the IP address into
           the output string. */
 
        else {
            sprintf( tqn, "%s", decode_addr(slots[i]->remote_ip_addr) ); }
 
        /* Now actually write it to the output file. */
 
        fprintf( fp, "%-6.3f  %-5d    %s\n", counts[i], rates[i], tqn );
        }
    fflush( fp );
    }

 
 
/*------------------------------------------------------------------------
--
--  dump_qrdata - a little routine to output a textual list of the
--  queryu/response records in  checker()'s database.
--
--  parameters:
--      file pointer, threshold
--
--  globals:
--      The query/response, ip address, and time queue tables are read.
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
float trate( cnt,ftim,ltim ) { return((float)cnt/(float)(ltim-ftim)); }
void dump_qrdata( fp )
FILE *fp;
    {
    u_long i, j, k, m, f, qr_total, lines;
    u_long ftime, tftime, ltime, tltime, resp;
    QR_HTABLE *qrhp;
    QR_DATA *qrp;
    IP_DATA *ip;
    char temp[100], tqn[100];
    QR_HTABLE *slots[QR_DUMP_SLOTS];
    u_long ccnt, counts[QR_DUMP_SLOTS];
    u_long ftimes[QR_DUMP_SLOTS], ltimes[QR_DUMP_SLOTS];
    time_t ttm;
    float pct, nsrate, qrrate;
 
    /* Get the current time. */
 
    ttm = time( (time_t *)NULL );
 
    /* Clear out the sort array */
 
    memset( slots, 0, sizeof(slots) );
    memset( counts, 0, sizeof(counts));
    memset( ltimes, 0, sizeof(ltimes));
    for( i=0; i<QR_DUMP_SLOTS; i++ ) ftimes[i] = 0xEFFFFFFF;

    /* Run through the entire QR hash queue, doing a sort
       on the access counts */
 
    for( i=0; i<QR_HASH_TABLE_SIZE; i++ ) {

        /* Loop through all of the entries at this hash slot */

        for( qrhp=qr_tbl[i]; qrhp != NULL; qrhp=qrhp->hclink ) {


            /* Loop through all of the query records for this name/type,
               one record for each client that has queried for that name. */

            for( ccnt=0,qrp=qrhp->qrlink; qrp != NULL; qrp=qrp->qrk.flink ) {
                if( qrp->qrq.q_first_time < tftime )
                    tftime = qrp->qrq.q_first_time;
                if( qrp->qrq.q_latest_time > tltime )
                    tltime = qrp->qrq.q_latest_time;
                ccnt += qrp->qrq.q_count; }

            /* Don't report on entries that only have one query, the
               rates and time spans make the calculations ugly. */

            if( ccnt == 1 ) continue;

            /* Look at the lowest entry on the sort list, if this one
               is greater than that one then find the place it should
               be inserted and move everything below it down one slot. */

            if( trate(ccnt,tftime,tltime) >
                trate(counts[QR_DUMP_SLOTS-1], ftimes[QR_DUMP_SLOTS-1],
                      ltimes[QR_DUMP_SLOTS-1]) ) {
                for( j=0; j<QR_DUMP_SLOTS-1; j++ ) {
                    if( trate(counts[j],ftimes[j],ltimes[j]) <
                        trate(ccnt,tftime,tltime) ) {
                        bcopy( &counts[j], &counts[j+1],
                            ((QR_DUMP_SLOTS-j)-1)*sizeof(u_long) );
                        bcopy( &slots[j], &slots[j+1],
                            ((QR_DUMP_SLOTS-j)-1)*sizeof(QR_HTABLE*) );
                        bcopy( &ftimes[j], &ftimes[j+1],
                            ((QR_DUMP_SLOTS-j)-1)*sizeof(u_long) );
                        bcopy( &ltimes[j], &ltimes[j+1],
                            ((QR_DUMP_SLOTS-j)-1)*sizeof(u_long) );
                        counts[j] = ccnt;
                        slots[j]  = qrhp;
                        ftimes[j] = tftime;
                        ltimes[j] = tltime;
                        break;
                        }
                    }
                }
            }
        }
 
    /* Now run through the resulting array and sum up the packet counts
       so that we can get a measure of how much of the traffic this
       list accounts for. */
 
    for( i=0,qr_total=0; i<QR_DUMP_SLOTS && slots[i]!=NULL; i++ ) {
         qr_total += counts[i]; }
 
    /* Dump out a heading for the query/response hash table listing */
 
    fprintf( fp, "\f\n" );
    fprintf( fp, "QUERY/RESPONSE TABLE at %s\n", ctime(&ttm) );
    fprintf( fp, "(list is %6.2f percent of total queries processed)\n",
        (float)(100.0*qr_total/total_queries) );
    fprintf( fp, "(list is %6.2f percent of query names in database)\n",
        (float)(100.0*i/total_qr_records) );
    fprintf( fp, "(qcnt is total query count, N is number of clients)\n" );
    fprintf( fp, "(rate is queries per hour for query name)\n" );
    fprintf( fp, "(pct is percent of all queries during active period)\n\n" );
    fprintf( fp, "-rate-- -N- -qcnt- --pct-- -rcode-- -type-- " );
    fprintf( fp, "---------query-name--------------\n" );
    fprintf( fp, "---------------------------------" );
    fprintf( fp, "-------------------------------------------\n" );
 
    /* Loop through all of the saved entries. */
 
    for( i=0; i<QR_DUMP_SLOTS && slots[i] != NULL; i++ ) {

        /* Set up a pointer, and then loop through all QR_DATA (one per
           client ip address) attached to this query name/type.  Keep
           track of the earliest and most recent accesses, and the number
           of different clients asking for this address. */

        qrhp = slots[i];
        ftime = 0xEFFFFFFF; ltime = 0; m = 0;
        for( qrp=qrhp->qrlink; qrp != NULL; qrp=qrp->qrk.flink ) { m++;
             if(qrp->qrq.q_first_time < ftime)  ftime=qrp->qrq.q_first_time;
             if(qrp->qrq.q_latest_time > ltime) {
                 resp = qrp->qrr.r_rcode;
                 ltime=qrp->qrq.q_latest_time; }
             }

        /* Calculate the query rate for this name and the percent of the
           nameserver's activity during the period that this name was
           active. */

        nsrate = calc_rate( ftime, ltime );
        if( ftime == ltime ) qrrate = 0.0;
        else qrrate = (float)counts[i] / (float)(ltime - ftime);
        pct = 100.0 * qrrate / nsrate;
        qrrate *= 3600.0;

        /* Create output strings for the query name, its type, and the
           most recent response code. */

        strncpy( temp, qrhp->q_name, sizeof(temp) );
        temp[47] = '\0';
        if( resp >= NUM_RCODE ) resp = NUM_RCODE-1;
        if( resp < 0 ) resp = NUM_RCODE-1;
        k = qrhp->q_type;
        if( k >= NUM_QTNAM ) k = NUM_QTNAM-1;
        if( k < 0 ) k = 0;

        /* Output the query name summary information. */

        fprintf( fp, "%6.2f  %-3d  %5d  %6.2f  %s  %s  %s\n", qrrate, m,
            counts[i], pct, rcode[resp], qtype_names[k], temp );
        fprintf( fp, "        " ); m = 8;

        /* Initialize the clients list output by sorting the query record
           linked list and setting our line count to zero. */
       
        sort_qr_list( qrhp );
        lines = 0;

        /* Output a list of clients asking for this name. */

        for( qrp=qrhp->qrlink; qrp != NULL; qrp=qrp->qrk.flink ) {

            /* only translate the name if it hasn't already been done. */

            ip = qrp->qrk.iprec;
            if( ip->resolved_name == NULL ){
                xlate_addr( ip->remote_ip_addr, temp, sizeof(temp) );
                temp[sizeof(temp)-1] = '\0';
                ip->resolved_name = (char *)calloc( 1, strlen(temp)+1 );
                strcpy( ip->resolved_name, temp );
                }

            /* If the new name string exceeds our page width then jump
               out of the loop, otherwise print it. */

            sprintf( tqn, "%s(%d), ", ip->resolved_name, qrp->qrq.q_count );
            if( m+strlen(tqn) > 78 ) {
                lines++;
                if( lines > 2 ) break;
                fprintf( fp, "\n        " );
                m = 8; }
            m += strlen(tqn);
            fprintf( fp, "%s", tqn );
            }

        /* If we had to bail out of that loop because we had too many
           clients to fit on one line then fill out the line with dots
           to show that fact. */

        if( qrp != NULL ) {
            memset( tqn, '.', sizeof(tqn) );
            tqn[80-m] = '\0';
            fprintf( fp, "%s\n\n", tqn ); }
        else fprintf( fp, "\n\n" );

        }
    fflush( fp );
    }

 
 
/*------------------------------------------------------------------------
--
--  dump_tq - a little routine to output a textual list of the
--  time deletion queue in checker()'s database.
--
--  parameters:
--      file pointer
--
--  globals:
--      The query/response, ip address, and time queue tables are read.
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void dump_tq( fp )
FILE *fp;
    {
    u_long i, j, f;
    QR_DATA *qrp;
    time_t ttm;
 
    /* Get the current time */
 
    ttm = time( (time_t *)NULL );
 
    /* Dump out a heading for the timed deletion queue contents. */
 
    fprintf( fp, "\n" );
    fprintf( fp, "TIME DELETION QUEUE at %s\n", ctime(&ttm) );
    fprintf( fp,
        "slot   ipaddr   type -------------query name--------------\n" );
    fprintf( fp,
        "----------------------------------------------------------\n" );
 
 
    /* Loop through the contents of the time deletion queue. */
 
    for( i=0; i<TIME_QUEUE_SLOT_COUNT; i++ ) {
        for( qrp=timeq[i].qrk.tq_flink; qrp!=NULL; qrp=qrp->qrk.tq_flink ) {
            fprintf( fp, "%04ld 0x%08lX %04ld %s\n", i,
                qrp->qrk.remote_ip_addr, (qrp->qrk.blink)->q_type,
                (qrp->qrk.blink)->q_name );
            }
        }
    fflush( fp );
    }

 
/*------------------------------------------------------------------------
--
--  signal handlers for SIGBUS and SIGSEGV
--
--  parameters:
--      none
--
--  globals:
        none
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
void sigbus_hdlr()
    {
    fprintf( ofp, "Report terminated due to SIGBUS\n" );
    fflush( ofp );
    finish_report();
    }
void sigsegv_hdlr()
    {
    fprintf( ofp, "Report terminated due to SIGSEGV\n" );
    fflush( ofp );
    finish_report();
    }

 
/*------------------------------------------------------------------------
--
--  Tell the parent process that the report is finished and then exit
--
--  parameters:
--      none
--
--  globals:
        none
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
finish_report()
    {
    char tstr[100];
 
    /* Notify the parent that we are finished. */
 
    strcpy( tstr, "done" );
    write( pipefd.out, tstr, strlen(tstr) );
    close( pipefd.out );
    close( pipefd.in );
    exit( );
    }
 

 
/*------------------------------------------------------------------------
--
--  Check that the report child process has finished
--
--  parameters:
--      none
--
--  globals:
--      none
--
--  return value:
--      none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
u_long check_report_done()
    {
    struct timeval tm;
    fd_set smask;
    char tstr[100];
    int rlen;
 
    /* Set up the select call that will check to see if there is anything
       available on the input file descriptor. */
 
    memset( &tm, 0, sizeof(tm) );
    FD_ZERO( &smask );
    FD_SET( pipefd.in, &smask );
 
    /* Check the file descriptor's contents, and read any message if 
       available. If the message says that the child is done then clean
       up the file descriptors. */
 
    if( select( FD_SETSIZE, &smask, NULL, NULL, &tm ) == 1 ) {
        rlen = read( pipefd.in, tstr, sizeof(tstr) );
        if( rlen < 0 || strncmp( tstr, "done", 4 ) == 0 ) {
            close( pipefd.in );
            close( pipefd.out );
            memset( &pipefd, 0, sizeof(pipefd) );
            return(1); }
        }
 
    return(0);
    }

 
/*------------------------------------------------------------------------
--
--  decode_eflag - dump out the strings corresponding to the eflag bits.
--
--  parameters:
--      standard error flag (u_long)
--
--  globals:
--      error string table
--
--  return value:
        none
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
decode_eflag( ef )
u_long ef;
    {
    u_long i,j;
 
    /* Just return if there were no errors, otherwise print
       the key token. */
 
    if( !ef ) return;
    printf( ":: " );
 
    /* Loop through all of the possible error bits. */
 
    for( i=0; i<NUMBER_OF_ERROR_TYPES; i++ ) {
        j = 1<<i;
        if( ef & j ) printf( "%s ", erstr[i] ); }
 
    /* Close out the output line (hope there weren't too many errors) */
 
    printf( "\n" );
    }

 
/*------------------------------------------------------------------------
--
--  decode_addr - translate a binary IP address into ascii
--
--  parameters:
--      IP address in binary
--
--  globals:
--      none
--
--  return value:
--      pointer to a string
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
char *decode_addr( ad )
u_long ad;
   {
   static char oadr[100];
   unsigned char tad[4];
 
   memcpy( tad, &ad, sizeof(tad) );
   sprintf( oadr, "%d.%d.%d.%d", tad[0], tad[1], tad[2], tad[3] );
   return( oadr );
   }

 
/*------------------------------------------------------------------------
--
--  calc_db_size() - a routine to estimate the amount of memory the
--  dynamically allocated database is using.
--
--  parameters:
--      none
--
--  globals:
--      record totals
--
--  return value:
--      database size in bytes
--
--  notes:
--      none
--
------------------------------------------------------------------------*/
u_long calc_db_size()
    {
    u_long calcd_db_size;
 
    /* Calculate the size of the database at this point. */
 
    calcd_db_size = (total_ht_records*sizeof(QR_HTABLE)) +
                    (total_qr_records*sizeof(QR_DATA)) +
                    (total_ip_records*sizeof(IP_DATA)) +
                    (total_er_records*sizeof(ERR_DATA)) +
                    (total_dm_records*sizeof(IPD_DATA)) +
                     total_qn_chars;
 
    return( calcd_db_size );
    }

 
/*--------------------------------------------------------------------------
--
--  xlate_addr()
--  This routine translates an IP address into a node name
--
--  parameters:
--      ipaddr - Internet address in 32bit binary form
--      nname  - pointer to string to hold domain name for input address
--
--  globals:
--      none
--
--  return value:
--      none
--
--  notes:
--      Uses dig program which may take a long time to return.
--
--------------------------------------------------------------------------*/
xlate_addr( ipaddr, nname, nsize )
unsigned long ipaddr;
char *nname;
int nsize;
    {
    FILE *fptr;
    char digstr[200], tok3[50], tname[300];
     
    /* Request the "dig" program to resolve the ip address into a
       domain node name. */
 
    remove( DIGFILE );
    sprintf( digstr, "%s/dig -x %s +rec | grep \'\tPTR\t\' > %s",
             dig_path, decode_addr(ipaddr), DIGFILE );
    system( digstr );
    usleep( 300000 );
            
    /* Clear out the name string. */
    
    nname[0] = '\0';
    tok3[0] = '\0';
 
    /* Open the output of the dig command, just return with a null
       string if something goes wrong. */
 
    if( (fptr = fopen( DIGFILE, "r" )) == NULL ) return;
 
    /* Parse the results of the "dig" command to get a node name. */
 
    while( 1 ) {
         if( fgets(digstr, sizeof(digstr), fptr) == NULL ) break;
         sscanf( digstr, "%*s %*s %s %s", tok3, tname );
         strncpy( nname, tname, nsize-1 );
         nname[nsize-1] = '\0';
         if( strcmp( tok3, "PTR" ) == 0 ) break;
         nname[0] = '\0';
         tok3[0] = '\0';
         }
 
    fclose( fptr );
    }

 
/*--------------------------------------------------------------------------
--
--  res_xlate_addr()
--  This routine translates an IP address into a node name
--
--  parameters:
--      ipaddr - Internet address in 32bit binary form
--      nname  - pointer to string to hold domain name for input address
--      namelen - length of buffer to put name into
--
--  globals:
--      none
--
--  return value:
--      0 - if the name could not be translated.
--      n - length of name that was returned if it was found      
--
--  notes:
--      Uses resolver query which may take a long time to return.
--
--------------------------------------------------------------------------*/
res_xlate_addr( ipaddr, nname, namlen )
unsigned long ipaddr;
char *nname;
int namlen;
    {
    static int inited=0;
    char *cp, *eom, *ptr;
    int i, s, n, rlen;
    HEADER *dh;
    char rbuf[PACKETSZ], inam[200];
    unsigned char adrstr[4];
    int rrtype, len;
 
    /* Initialize the resolver state variables if they have not already
       been initialized; and clear out our message buffer. */
 
    if( !inited ) { res_init(); inited=1; }
    memset( rbuf, 0, sizeof(rbuf) );
 
    /* Create the query from the incoming IP address */
 
    memcpy( adrstr, &ipaddr, sizeof(unsigned long) );
    sprintf( inam, "%d.%d.%d.%d.in-addr.arpa", adrstr[3], adrstr[2],
       adrstr[1], adrstr[0] );
 
    /* Issue the name resolution query to the nameserver daemon.  If an
       error is encountered then return as if no match had been found. */
 
    rlen = res_search( inam, C_IN, T_ANY, rbuf, sizeof(rbuf) );
    if( rlen < 0 ) { herror("res_search"); return(0); }
 
    /* Set up some pointers into the returned response packet. */
 
    dh = (HEADER *)rbuf;
    cp = rbuf + sizeof(HEADER);
    eom = rbuf + rlen;
 
    /*-------------------------------------------------------------------
    -- PARSE THE RESPONSE PACKET
    --------------------------------------------------------------------*/
 
    /* Skip the question record holding the name we sent up to the
       nameserver. */
 
    n = dn_skipname( cp, eom );
    if( n < 0 ) { herror("dn_skipname 1"); return(0); }
    cp += (n + sizeof(short) + sizeof(short));
 
    /* Loop through the answer records looking for address or text data */
 
    for( i=0; i<(int)dh->ancount; i++ )
        {
 
        /* Skip the name field, should be the same as what we send up. */
 
        n = dn_skipname( cp, eom );
        if( n < 0 ) { herror("dn_skipname 2"); return(0); }
        cp += n;
 
        /* Get the resource record type and skip the class, and ttl */
 
        if( (cp+1) > eom ) { return(0); }
        rrtype = _getshort( cp );
        cp += (sizeof(short) + sizeof(short) + sizeof(long));
 
        /* Get the length of the resource data area. */
 
        if( (cp+1) > eom ) { return(0); }
        rlen = _getshort( cp );
        cp += sizeof(short);
 
        /* If the resource record is a pointer record then it probably has
           the information we are looking for. */
 
        if( rrtype == T_PTR ) {
            if( (cp+rlen) > eom ) { return(0); }
            n = dn_expand( rbuf, eom, cp, nname, namlen );
            if( n < 0 ) { herror("dn_expand"); return(0); }
            return( n );
            }
 
        /* Make sure to skip the pointer ahead to account for the data. */
 
        cp += rlen;
        }
     }
 
 

